"""This module contains CUDA source code for the evaluation of the coupling
matrix from lookups."""

# The following cuda kernel multiplies the coupling matrix to a vector. It is based on linear interpolation of
# the lookup table.
#
# input arguments of the cuda kernel:
# n (np.uint32):  n[i] contains the mutlipole multi-index with regard to self.l_max and self.m_max of the i-th
#                 entry of a system vector
# m (np.float32): m[i] contains the multipole order with regard to particle.l_max and particle.m_max
# x_pos (np.float32): x_pos[i] contains the respective particle x-position
# y_pos (np.float32): y_pos[i] contains the respective particle y-position
# z_pos (np.float32): z_pos[i] contains the respective particle z-position
# re_lookup_pl (np.float32): the real part of the lookup table for the z1+z2 part of the Sommerfeld integral,
#                            in the format (rho, sum_z, n1, n2)
# im_lookup_pl (np.float32): the imaginary part of the lookup table for the z1+z2 part of the Sommerfeld
#                            integral, in the format (rho, sum_z, n1, n2)
# re_lookup_mn (np.float32): the real part of the lookup table for the z1-z2 part of the Sommerfeld integral,
#                            in the format (rho, diff_z, n1, n2)
# im_lookup_mn (np.float32): the imaginary part of the lookup table for the z1-z2 part of the Sommerfeld
#                            integral, in the format (rho, diff_z, n1, n2)
# re_in_vec (np.float32): the real part of the vector to be multiplied with the coupling matrix
# im_in_vec (np.float32): the imaginary part of the vector to be multiplied with the coupling matrix
# re_result_vec (np.float32): the real part of the vector into which the result is written
# im_result_vec (np.float32): the imaginary part of the vector into which the result is written
linear_volume_lookup_source = """
    #define BLOCKSIZE %i
    #define NUMBER_OF_UNKNOWNS %i
    #define Z_ARRAY_LENGTH %i
    #define MIN_RHO %f
    #define MIN_Z_SUM %f
    #define MIN_Z_DIFF %f
    #define LOOKUP_RESOLUTION %f
    __global__ void coupling_kernel(const int *n, const float *m, const float *x_pos, const float *y_pos,
                                    const float *z_pos, const float *re_lookup_pl, const float *im_lookup_pl,
                                    const float *re_lookup_mn, const float *im_lookup_mn,
                                    const float *re_in_vec, const float *im_in_vec,
                                    float  *re_result, float  *im_result)
    {
        unsigned int i1 = blockIdx.x * blockDim.x + threadIdx.x;
        if(i1 >= NUMBER_OF_UNKNOWNS) return;

        const float x1 = x_pos[i1];
        const float y1 = y_pos[i1];
        const float z1 = z_pos[i1];

        const int n1 = n[i1];
        const float m1 = m[i1];

        re_result[i1] = 0.0;
        im_result[i1] = 0.0;

        for (int i2=0; i2<NUMBER_OF_UNKNOWNS; i2++)
        {
            float x21 = x1 - x_pos[i2];
            float y21 = y1 - y_pos[i2];
            float sz21 = z1 + z_pos[i2];
            float dz21 = z1 - z_pos[i2];

            const int n2 = n[i2];
            const float m2 = m[i2];

            float rho = sqrt(x21*x21+y21*y21);
            float phi = atan2(y21,x21);

            int rho_idx = (int) floor((rho - MIN_RHO) / LOOKUP_RESOLUTION);
            float rho_w = (rho - MIN_RHO) / LOOKUP_RESOLUTION - floor((rho - MIN_RHO) / LOOKUP_RESOLUTION);

            int sz_idx = (int) floor((sz21 - MIN_Z_SUM) / LOOKUP_RESOLUTION);
            float sz_w = (sz21 - MIN_Z_SUM) / LOOKUP_RESOLUTION - floor((sz21 - MIN_Z_SUM) / LOOKUP_RESOLUTION);

            int dz_idx = (int) floor((dz21 - MIN_Z_DIFF) / LOOKUP_RESOLUTION);
            float dz_w = (dz21 - MIN_Z_DIFF) / LOOKUP_RESOLUTION 
                            - floor((dz21 - MIN_Z_DIFF) /  LOOKUP_RESOLUTION);

            int idx_rho_sz = rho_idx * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE 
                                +  sz_idx * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            int idx_rho_szpl1 = rho_idx * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE
                                + (sz_idx + 1) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            int idx_rhopl1_sz = (rho_idx + 1) * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE
                                + sz_idx * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            int idx_rhopl1_szpl1 = (rho_idx + 1) * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE
                                    + (sz_idx + 1) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            int idx_rho_dz = rho_idx * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE + dz_idx * BLOCKSIZE * BLOCKSIZE
                                + n1 * BLOCKSIZE + n2;

            int idx_rho_dzpl1 = rho_idx * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE
                                + (dz_idx + 1) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            int idx_rhopl1_dz = (rho_idx + 1) * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE
                                + dz_idx * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            int idx_rhopl1_dzpl1 = (rho_idx + 1) * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE
                                    + (dz_idx + 1) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            float f_rhoi = re_lookup_pl[idx_rho_sz] * (1 - sz_w) + re_lookup_pl[idx_rho_szpl1] * sz_w;
            float f_rhoipl1 = re_lookup_pl[idx_rhopl1_sz] * (1 - sz_w) + re_lookup_pl[idx_rhopl1_szpl1] * sz_w;
            float re_si_pl = f_rhoi * (1 - rho_w) + f_rhoipl1 * rho_w;

            f_rhoi = im_lookup_pl[idx_rho_sz] * (1 - sz_w) + im_lookup_pl[idx_rho_szpl1] * sz_w;
            f_rhoipl1 = im_lookup_pl[idx_rhopl1_sz] * (1 - sz_w) + im_lookup_pl[idx_rhopl1_szpl1] * sz_w;
            float im_si_pl = f_rhoi * (1 - rho_w) + f_rhoipl1 * rho_w;

            f_rhoi = re_lookup_mn[idx_rho_dz] * (1 - dz_w) + re_lookup_mn[idx_rho_dzpl1] * dz_w;
            f_rhoipl1 = re_lookup_mn[idx_rhopl1_dz] * (1 - dz_w) + re_lookup_mn[idx_rhopl1_dzpl1] * dz_w;
            float re_si_mn = f_rhoi * (1 - rho_w) + f_rhoipl1 * rho_w;

            f_rhoi = im_lookup_mn[idx_rho_dz] * (1 - dz_w) + im_lookup_mn[idx_rho_dzpl1] * dz_w;
            f_rhoipl1 = im_lookup_mn[idx_rhopl1_dz] * (1 - dz_w) + im_lookup_mn[idx_rhopl1_dzpl1] * dz_w;
            float im_si_mn = f_rhoi * (1 - rho_w) + f_rhoipl1 * rho_w;

            float re_eimphi = cosf((m2 - m1) * phi);
            float im_eimphi = sinf((m2 - m1) * phi);

            float re_w = re_eimphi * (re_si_pl + re_si_mn) - im_eimphi * (im_si_pl + im_si_mn);
            float im_w = im_eimphi * (re_si_pl + re_si_mn) + re_eimphi * (im_si_pl + im_si_mn);

            re_result[i1] += re_w * re_in_vec[i2] - im_w * im_in_vec[i2];
            im_result[i1] += re_w * im_in_vec[i2] + im_w * re_in_vec[i2];
        }
    }"""

# The following cuda kernel multiplies the coupling matrix to a vector. It is based on cubic Hermite spline interpolation
# of the lookup table.
#
# input arguments of the cuda kernel:
# n (np.uint32):  n[i] contains the mutlipole multi-index with regard to self.l_max and self.m_max of the i-th
#                 entry of a system vector
# m (np.float32): m[i] contains the multipole order with regard to particle.l_max and particle.m_max
# x_pos (np.float32): x_pos[i] contains the respective particle x-position
# y_pos (np.float32): y_pos[i] contains the respective particle y-position
# z_pos (np.float32): z_pos[i] contains the respective particle z-position
# re_lookup_pl (np.float32): the real part of the lookup table for the z1+z2 part of the Sommerfeld integral,
#                            in the format (rho, sum_z, n1, n2)
# im_lookup_pl (np.float32): the imaginary part of the lookup table for the z1+z2 part of the Sommerfeld
#                            integral, in the format (rho, sum_z, n1, n2)
# re_lookup_mn (np.float32): the real part of the lookup table for the z1-z2 part of the Sommerfeld integral,
#                            in the format (rho, diff_z, n1, n2)
# im_lookup_mn (np.float32): the imaginary part of the lookup table for the z1-z2 part of the Sommerfeld
#                            integral, in the format (rho, diff_z, n1, n2)
# re_in_vec (np.float32): the real part of the vector to be multiplied with the coupling matrix
# im_in_vec (np.float32): the imaginary part of the vector to be multiplied with the coupling matrix
# re_result_vec (np.float32): the real part of the vector into which the result is written
# im_result_vec (np.float32): the imaginary part of the vector into which the result is written
cubic_volume_lookup_source = """
    #define BLOCKSIZE %i
    #define NUMBER_OF_UNKNOWNS %i
    #define Z_ARRAY_LENGTH %i
    #define MIN_RHO %f
    #define MIN_Z_SUM %f
    #define MIN_Z_DIFF %f
    #define LOOKUP_RESOLUTION %f

    __device__ int lookup_index(int const i_r, int const i_z, int const n1, int const n2)
    {
        return i_r * Z_ARRAY_LENGTH * BLOCKSIZE * BLOCKSIZE + i_z * BLOCKSIZE * BLOCKSIZE + n1 *  BLOCKSIZE + n2;
    }

    __device__ float cubic_interpolation(float const w, float const lookup_imn1, float const lookup_i, 
                                            float const lookup_ipl1, float const lookup_ipl2)
    {
        return ((-w*w*w+2*w*w-w) * lookup_imn1 + (3*w*w*w-5*w*w+2) * lookup_i + (-3*w*w*w+4*w*w+w) * lookup_ipl1
                + (w*w*w-w*w) * lookup_ipl2) / 2;
    }

    __device__ float lookup_2D(int const i_rho, float const w_rho, int const i_z, float const w_z, int const n1,
                                int const n2, float const *lookup)
    {
        int i_zmn1 = lookup_index(i_rho-1, i_z-1, n1, n2);
        int i = lookup_index(i_rho-1, i_z, n1, n2);
        int i_zpl1 = lookup_index(i_rho-1, i_z+1, n1, n2);
        int i_zpl2 = lookup_index(i_rho-1, i_z+2, n1, n2);
        float lookup_rmn1 = cubic_interpolation(w_z, lookup[i_zmn1], lookup[i], lookup[i_zpl1], lookup[i_zpl2]);

        i_zmn1 = lookup_index(i_rho, i_z-1, n1, n2);
        i = lookup_index(i_rho, i_z, n1, n2);
        i_zpl1 = lookup_index(i_rho, i_z+1, n1, n2);
        i_zpl2 = lookup_index(i_rho, i_z+2, n1, n2);
        float lookup_r = cubic_interpolation(w_z, lookup[i_zmn1], lookup[i], lookup[i_zpl1], lookup[i_zpl2]);

        i_zmn1 = lookup_index(i_rho+1, i_z-1, n1, n2);
        i = lookup_index(i_rho+1, i_z, n1, n2);
        i_zpl1 = lookup_index(i_rho+1, i_z+1, n1, n2);
        i_zpl2 = lookup_index(i_rho+1, i_z+2, n1, n2);
        float lookup_rpl1 = cubic_interpolation(w_z, lookup[i_zmn1], lookup[i], lookup[i_zpl1], lookup[i_zpl2]);

        i_zmn1 = lookup_index(i_rho+2, i_z-1, n1, n2);
        i = lookup_index(i_rho+2, i_z, n1, n2);
        i_zpl1 = lookup_index(i_rho+2, i_z+1, n1, n2);
        i_zpl2 = lookup_index(i_rho+2, i_z+2, n1, n2);
        float lookup_rpl2 = cubic_interpolation(w_z, lookup[i_zmn1], lookup[i], lookup[i_zpl1], lookup[i_zpl2]);

        return cubic_interpolation(w_rho, lookup_rmn1, lookup_r, lookup_rpl1, lookup_rpl2);
    }

    __global__ void coupling_kernel(const int *n, const float *m, const float *x_pos, const float *y_pos,
                                    const float *z_pos, const float *re_lookup_pl, const float *im_lookup_pl,
                                    const float *re_lookup_mn, const float *im_lookup_mn,
                                    const float *re_in_vec, const float *im_in_vec,
                                    float  *re_result, float  *im_result)
    {
        unsigned int i1 = blockIdx.x * blockDim.x + threadIdx.x;
        if(i1 >= NUMBER_OF_UNKNOWNS) return;

        const float x1 = x_pos[i1];
        const float y1 = y_pos[i1];
        const float z1 = z_pos[i1];

        const int n1 = n[i1];
        const float m1 = m[i1];

        re_result[i1] = 0.0;
        im_result[i1] = 0.0;

        for (int i2=0; i2<NUMBER_OF_UNKNOWNS; i2++)
        {
            float x21 = x1 - x_pos[i2];
            float y21 = y1 - y_pos[i2];
            float sz21 = z1 + z_pos[i2];
            float dz21 = z1 - z_pos[i2];

            const int n2 = n[i2];
            const float m2 = m[i2];

            float rho = sqrt(x21*x21+y21*y21);
            float phi = atan2(y21,x21);

            int rho_idx = (int) floor((rho - MIN_RHO) / LOOKUP_RESOLUTION);
            float rho_w = (rho - MIN_RHO) / LOOKUP_RESOLUTION - floor((rho - MIN_RHO) / LOOKUP_RESOLUTION);

            int sz_idx = (int) floor((sz21 - MIN_Z_SUM) / LOOKUP_RESOLUTION);
            float sz_w = (sz21 - MIN_Z_SUM) / LOOKUP_RESOLUTION - floor((sz21 - MIN_Z_SUM) / LOOKUP_RESOLUTION);

            int dz_idx = (int) floor((dz21 - MIN_Z_DIFF) / LOOKUP_RESOLUTION);
            float dz_w = (dz21 - MIN_Z_DIFF) / LOOKUP_RESOLUTION 
                            - floor((dz21 - MIN_Z_DIFF) /  LOOKUP_RESOLUTION);

            float re_si_pl = lookup_2D(rho_idx, rho_w, sz_idx, sz_w, n1, n2, re_lookup_pl);
            float im_si_pl = lookup_2D(rho_idx, rho_w, sz_idx, sz_w, n1, n2, im_lookup_pl);
            float re_si_mn = lookup_2D(rho_idx, rho_w, dz_idx, dz_w, n1, n2, re_lookup_mn);
            float im_si_mn = lookup_2D(rho_idx, rho_w, dz_idx, dz_w, n1, n2, im_lookup_mn);

            float re_eimphi = cosf((m2 - m1) * phi);
            float im_eimphi = sinf((m2 - m1) * phi);

            float re_w = re_eimphi * (re_si_pl + re_si_mn) - im_eimphi * (im_si_pl + im_si_mn);
            float im_w = im_eimphi * (re_si_pl + re_si_mn) + re_eimphi * (im_si_pl + im_si_mn);

            re_result[i1] += re_w * re_in_vec[i2] - im_w * im_in_vec[i2];
            im_result[i1] += re_w * im_in_vec[i2] + im_w * re_in_vec[i2];
        }
    }"""

# This cuda kernel multiplies the coupling matrix to a vector. It is based on linear interpolation of the lookup
# table.
#
# input arguments of the cuda kernel:
# n (np.uint32):  n[i] contains the mutlipole multi-index with regard to self.l_max and self.m_max of the i-th
#                 entry of a system vector
# m (np.float32): m[i] contains the multipole order with regard to particle.l_max and particle.m_max
# x_pos (np.float32): x_pos[i] contains the respective particle x-position
# y_pos (np.float32): y_pos[i] contains the respective particle y-position
# re_lookup (np.float32): the real part of the lookup table, in the format [r, n1, n2]
# im_lookup (np.float32): the imaginary part of the lookup table, in the format [r, n1, n2]
# re_in_vec (np.float32): the real part of the vector to be multiplied with the coupling matrix
# im_in_vec (np.float32): the imaginary part of the vector to be multiplied with the coupling matrix
# re_result_vec (np.float32): the real part of the vector into which the result is written
# im_result_vec (np.float32): the imaginary part of the vector into which the result is written
linear_radial_lookup_source = """
    #define BLOCKSIZE %i
    #define NUMBER_OF_UNKNOWNS %i
    #define MIN_RHO %f
    #define LOOKUP_RESOLUTION %f

    __global__ void coupling_kernel(const int *n, const float *m, const float *x_pos, const float *y_pos,
                                    const float *re_lookup, const float *im_lookup, const float *re_in_vec,
                                    const float *im_in_vec, float  *re_result, float  *im_result)
    {
        unsigned int i1 = blockIdx.x * blockDim.x + threadIdx.x;
        if(i1 >= NUMBER_OF_UNKNOWNS) return;

        const float x1 = x_pos[i1];
        const float y1 = y_pos[i1];

        const int n1 = n[i1];
        const float m1 = m[i1];

        re_result[i1] = 0.0;
        im_result[i1] = 0.0;

        for (int i2=0; i2<NUMBER_OF_UNKNOWNS; i2++)
        {
            float x21 = x1 - x_pos[i2];
            float y21 = y1 - y_pos[i2];

            const int n2 = n[i2];
            const float m2 = m[i2];

            float r = sqrt(x21*x21+y21*y21);
            float phi = atan2(y21,x21);

            int r_idx = (int) floor((r - MIN_RHO) / LOOKUP_RESOLUTION);
            float w = (r - MIN_RHO) / LOOKUP_RESOLUTION - floor((r - MIN_RHO) / LOOKUP_RESOLUTION);

            int idx = r_idx * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;
            int idx_pl_1 = (r_idx + 1) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            float re_si = re_lookup[idx] + w * (re_lookup[idx_pl_1] - re_lookup[idx]);
            float im_si = im_lookup[idx] + w * (im_lookup[idx_pl_1] - im_lookup[idx]);

            float re_eimphi = cosf((m2 - m1) * phi);
            float im_eimphi = sinf((m2 - m1) * phi);

            float re_w = re_eimphi * re_si - im_eimphi * im_si;
            float im_w = im_eimphi * re_si + re_eimphi * im_si;

            re_result[i1] += re_w * re_in_vec[i2] - im_w * im_in_vec[i2];
            im_result[i1] += re_w * im_in_vec[i2] + im_w * re_in_vec[i2];

        }
    }"""

# This cuda kernel multiplies the coupling matrix to a vector. It is based on cubic Hermite spline interpolation of the
# lookup table.
#
# input arguments of the cuda kernel:
# n (np.uint32):  n[i] contains the mutlipole multi-index with regard to self.l_max and self.m_max of the i-th
#                 entry of a system vector
# m (np.float32): m[i] contains the multipole order with regard to particle.l_max and particle.m_max
# x_pos (np.float32): x_pos[i] contains the respective particle x-position
# y_pos (np.float32): y_pos[i] contains the respective particle y-position
# re_lookup (np.float32): the real part of the lookup table, in the format [r, n1, n2]
# im_lookup (np.float32): the imaginary part of the lookup table, in the format [r, n1, n2]
# re_in_vec (np.float32): the real part of the vector to be multiplied with the coupling matrix
# im_in_vec (np.float32): the imaginary part of the vector to be multiplied with the coupling matrix
# re_result_vec (np.float32): the real part of the vector into which the result is written
# im_result_vec (np.float32): the imaginary part of the vector into which the result is written
cubic_radial_lookup_source = """
    #define BLOCKSIZE %i
    #define NUMBER_OF_UNKNOWNS %i
    #define MIN_RHO %f
    #define LOOKUP_RESOLUTION %f

    __device__ float cubic_interpolation(float const w, float const lookup_imn1, float const lookup_i,
                                         float const lookup_ipl1, float const lookup_ipl2)
    {
        return ((-w*w*w+2*w*w-w) * lookup_imn1 + (3*w*w*w-5*w*w+2) * lookup_i + (-3*w*w*w+4*w*w+w) * lookup_ipl1
                + (w*w*w-w*w) * lookup_ipl2) / 2;
    }


    __global__ void coupling_kernel(const int *n, const float *m, const float *x_pos, const float *y_pos,
                                    const float *re_lookup, const float *im_lookup, const float *re_in_vec,
                                    const float *im_in_vec, float  *re_result, float  *im_result)
    {
        unsigned int i1 = blockIdx.x * blockDim.x + threadIdx.x;
        if(i1 >= NUMBER_OF_UNKNOWNS) return;

        const float x1 = x_pos[i1];
        const float y1 = y_pos[i1];

        const int n1 = n[i1];
        const float m1 = m[i1];

        re_result[i1] = 0.0;
        im_result[i1] = 0.0;

        for (int i2=0; i2<NUMBER_OF_UNKNOWNS; i2++)
        {
            float x21 = x1 - x_pos[i2];
            float y21 = y1 - y_pos[i2];

            const int n2 = n[i2];
            const float m2 = m[i2];

            float r = sqrt(x21*x21+y21*y21);
            float phi = atan2(y21,x21);

            int r_idx = (int) floor((r - MIN_RHO) / LOOKUP_RESOLUTION);
            float w = (r - MIN_RHO) / LOOKUP_RESOLUTION - floor((r - MIN_RHO) / LOOKUP_RESOLUTION);

            int idx_mn1 = (r_idx - 1) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;
            int idx = r_idx * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;
            int idx_pl1 = (r_idx + 1) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;
            int idx_pl2 = (r_idx + 2) * BLOCKSIZE * BLOCKSIZE + n1 * BLOCKSIZE + n2;

            float re_si = cubic_interpolation(w, re_lookup[idx_mn1], re_lookup[idx], re_lookup[idx_pl1], 
                                              re_lookup[idx_pl2]);
            float im_si = cubic_interpolation(w, im_lookup[idx_mn1], im_lookup[idx], im_lookup[idx_pl1], 
                                              im_lookup[idx_pl2]);                                                

            float re_eimphi = cosf((m2 - m1) * phi);
            float im_eimphi = sinf((m2 - m1) * phi);

            float re_w = re_eimphi * re_si - im_eimphi * im_si;
            float im_w = im_eimphi * re_si + re_eimphi * im_si;

            re_result[i1] += re_w * re_in_vec[i2] - im_w * im_in_vec[i2];
            im_result[i1] += re_w * im_in_vec[i2] + im_w * re_in_vec[i2];
        }
    }"""



"""This module contains several mathematical functions."""

# -*- coding: utf-8 -*-
import numpy as np
import scipy.special
import warnings
import smuthi.utility.memoizing as memo
import sys
import sympy
import math
import smuthi.utility.numba_helpers as nh
from numba import jit
from sympy.physics.quantum.spin import Rotation
try:
    try:
        from numba.core.typing import cffi_utils
    except ModuleNotFoundError: # typically when numba.__version__ <= '0.48.0'
        from numba.typing import cffi_utils
    from pywigxjpf_ffi import ffi, lib
    import pywigxjpf_ffi
    cffi_utils.register_module(pywigxjpf_ffi)
    nb_wig3jj = pywigxjpf_ffi.lib.wig3jj

    lib.wig_table_init(100,9)
    lib.wig_temp_init(100)
except Exception as e:
    sys.stdout.write(
        'pywigxjpf import failed with the following error message: \n"'
        + str(e)
        + '"\nUsing sympy implementaion of Wigner-3j symbols instead.\n'
          'In certain cases, this can significantly increase the simulation time.\n')
    sys.stdout.flush()
    from sympy.physics.wigner import wigner_3j
    def nb_wig3jj(jj_1, jj_2, jj_3, mm_1, mm_2, mm_3):
        return float(wigner_3j(jj_1/2, jj_2/2, jj_3/2, mm_1/2, mm_2/2, mm_3/2))

    
def legendre_normalized(ct, st, lmax):
    r"""Return the normalized associated Legendre function :math:`P_l^m(\cos\theta)` and the angular functions
    :math:`\pi_l^m(\cos \theta)` and :math:`\tau_l^m(\cos \theta)`, as defined in
    `A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006
    <https://doi.org/10.1007/978-3-540-33697-6>`_.
    Two arguments (ct and st) are passed such that the function is valid for general complex arguments, while the branch
    cuts are defined by the user already in the definition of st.

    Args:
        ct (ndarray): cosine of theta (or kz/k)
        st (ndarray): sine of theta (or kp/k), need to have same dimension as ct, and st**2+ct**2=1 is assumed
        lmax (int): maximal multipole order

    Returns:
        - ndarray plm[l, m, *ct.shape] contains :math:`P_l^m(\cos \theta)`. The entries of the list have same dimension as ct (and st)
        - ndarray pilm[l, m, *ct.shape] contains :math:`\pi_l^m(\cos \theta)`.
        - ndarray taulm[l, m, *ct.shape] contains :math:`\tau_l^m(\cos \theta)`.
    """
    if hasattr(ct, '__len__'):
        ct = np.array(ct, dtype=np.complex128)
    else:
        ct = np.array([ct], dtype=np.complex128)

    if hasattr(st, '__len__'):
        st = np.array(st, dtype=np.complex128)
    else:
        st = np.array([st], dtype=np.complex128)

    return legendre_normalized_numbed(ct, st, lmax)

@jit(nopython=True, cache=True, nogil=True)
def legendre_normalized_numbed(ct, st, lmax):
    plm = np.zeros((lmax+1, lmax+1, *ct.shape), dtype=np.complex128)
    pilm = np.zeros((lmax+1, lmax+1, *ct.shape), dtype=np.complex128)
    taulm = np.zeros((lmax+1, lmax+1, *ct.shape), dtype=np.complex128)
    pprimel0 = np.zeros((lmax+1, *ct.shape), dtype=np.complex128)

    plm[0,0] = np.sqrt(2)/2
    plm[1, 0] = np.sqrt(3/2) * ct
    pprimel0[1] = np.sqrt(3) * plm[0, 0]
    taulm[0, 0] = -st * pprimel0[0]
    taulm[1, 0] = -st * pprimel0[1]

    for l in range(1, lmax):
        plm[l + 1, 0] = (1 / (l + 1) * np.sqrt((2 * l + 1) * (2 * l + 3)) * ct * plm[l, 0] -
                         l / (l + 1) * np.sqrt((2 * l + 3) / (2 * l - 1)) * plm[l-1, 0])
        pprimel0[l + 1] = ((l + 1) * np.sqrt((2 * (l + 1) + 1) / (2 * (l + 1) - 1)) * plm[l, 0] +
                           np.sqrt((2 * (l + 1) + 1) / (2 * (l + 1) - 1)) * ct * pprimel0[l])
        taulm[l + 1, 0] = -st * pprimel0[l + 1]

    for m in range(1, lmax + 1):
        prefactor = nh.prefactor_expansion(m)
        plm[m, m] = prefactor * st**m
        pilm[m, m] = prefactor * st**(m - 1)
        taulm[m, m] = m * ct * pilm[m, m]
        for l in range(m, lmax):
            plm[l + 1, m] = (np.sqrt((2 * l + 1) * (2 * l + 3) / ((l + 1 - m) * (l + 1 + m))) * ct * plm[l, m] -
                             np.sqrt((2 * l + 3) * (l - m) * (l + m) / ((2 * l - 1) * (l + 1 - m) * (l + 1 + m))) *
                             plm[l - 1, m])
            pilm[l + 1, m] = (np.sqrt((2 * l + 1) * (2 * l + 3) / (l + 1 - m) / (l + 1 + m)) * ct * pilm[l, m] -
                              np.sqrt((2 * l + 3) * (l - m) * (l + m) / (2 * l - 1) / (l + 1 - m) / (l + 1 + m)) *
                              pilm[l - 1, m])
            taulm[l + 1, m] = ((l + 1) * ct * pilm[l + 1, m] -
                               (l + 1 + m) * np.sqrt((2 * (l + 1) + 1) * (l + 1 - m) / (2 * (l + 1) - 1) / (l + 1 + m))
                               * pilm[l, m])

    return plm, pilm, taulm


spherical_bessel = scipy.special.spherical_jn


def spherical_hankel(n, x):
    spherj = scipy.special.spherical_jn(n, x)
    sphery = scipy.special.spherical_yn(n, x)
    if hasattr(x, '__len__'):
        sphery[x==0] = np.nan
    return spherj + 1j * sphery


def dx_xj(n, x):
    r"""Derivative of :math:`x j_n(x)`, where :math:`j_n(x)` is the spherical Bessel function.

    Args:
        n (int): (n>0): Order of spherical Bessel function
        x (array, complex or float): Argument for spherical Bessel function

    Returns:
        Derivative :math:`\partial_x(x j_n(x))` as array.
    """
    res = x * spherical_bessel(n - 1, x) - n * spherical_bessel(n, x)
    return res


def dx_xh(n, x):
    r"""Derivative of :math:`x h_n(x)`, where :math:`h_n(x)` is the spherical Hankel function.

    Args:
        n (int): (n>0): Order of spherical Bessel function
        x (array, complex or float): Argument for spherical Hankel function

    Returns:
        Derivative :math:`\partial_x(x h_n(x))` as array.
    """
    res = x * spherical_hankel(n - 1, x) - n * spherical_hankel(n, x)
    return res


@memo.Memoize
def factorial(n):
    """Return factorial.

    Args:
        n (int): Argument (non-negative)

    Returns:
        Factorial of n
    """
    assert type(n) == int and n >= 0
    if n == 0:
        return 1
    else:
        return n * factorial(n-1)


@memo.Memoize
def legendre_prefactor(m):
    """Returns the prefactor :math:`\sqrt(\frac{(2*m+1)}{2(2m)!} (2m-1)!!`
    taking advantage of python's bignumbers.
    A jitted version of this function is available as jitted_prefactor() under
    numba_helpers

    Args:
        m (int64): Argument (non-negative)

    Returns:
        :math:`\sqrt(\frac{(2*m+1)!!}{2(2m)!!}`
    """
    res = 1.
    for t in range(2,2*m+2,2):
        res += res/t # equivalent to res *= (t+1)/t, but retains more significant digits
    return (res/2)**0.5


def wigner_d(l, m, m_prime, beta, wdsympy=False):
    """Computation of Wigner-d-functions for the rotation of a T-matrix

    Args:
        l (int):          Degree :math:`l` (1, ..., lmax)
        m (int):          Order :math:`m` (-min(l,mmax),...,min(l,mmax))
        m_prime (int):    Order :math:`m_prime` (-min(l,mmax),...,min(l,mmax))
        beta (float):     Second Euler angle in rad
        wdsympy (bool):   If True, Wigner-d-functions come from the sympy toolbox 
        
    Returns:
        real value of Wigner-d-function
    """
    wig_d = np.zeros(l + 1, dtype=complex)
    
    if wdsympy == False:
    
        if beta < 0:
            aa = m
            bb = m_prime
            m = bb
            m_prime = aa
           
        if m == 0 and m_prime == 0:
            for nn in range(1, l + 1):
                wig_d[nn] = sympy.legendre_poly(nn, np.cos(beta))          
        else:
            # recursion formulation (Mishchenko, Scattering, Absorption and Emission of Light by small Particles, p.365 (B.22 - B.24))
            l_min = max(abs(m), abs(m_prime))
            wig_d[l_min - 1] = 0
            if m_prime >= m:
                zeta = 1
            else:
                zeta = (-1) ** (m - m_prime)
        
            wig_d[l_min] = (zeta * 2.0 ** (-l_min) * (factorial(2 * l_min) / (factorial(abs(m - m_prime)) 
                                                                              * factorial(abs(m + m_prime)))) ** 0.5
                            * (1 - np.cos(beta)) ** (abs(m - m_prime) / 2) 
                            * (1 + np.cos(beta)) ** (abs(m + m_prime) / 2 ))
    
            for ll in range(l_min, l):
                wig_d[ll + 1] = (((2 * ll + 1) * (ll * (ll + 1) * np.cos(beta) - m * m_prime) * wig_d[ll] 
                                - (ll + 1) * (ll ** 2 - m ** 2) ** 0.5 * (ll ** 2 - m_prime ** 2) ** 0.5 
                                * wig_d[ll - 1]) / (ll * ((ll + 1) ** 2 - m ** 2) ** 0.5 
                                                    * ((ll + 1) ** 2 - m_prime ** 2) ** 0.5))
    
    else:
        wig_d[l] = complex(Rotation.d(l, m, m_prime, beta).doit())
      
    return wig_d[l].real


def wigner_D(l , m, m_prime, alpha, beta, gamma, wdsympy=False):
    """Computation of Wigner-D-functions for the rotation of a T-matrix
         
    Args:
        l (int):          Degree :math:`l` (1, ..., lmax)
        m (int):          Order :math:`m` (-min(l,mmax),...,min(l,mmax))
        m_prime (int):    Order :math:`m_prime` (-min(l,mmax),...,min(l,mmax))
        alpha (float):    First Euler angle in rad
        beta (float):     Second Euler angle in rad
        gamma (float):    Third Euler angle in rad
        wdsympy (bool):   If True, Wigner-d-functions come from the sympy toolbox
        
    Returns:
        single complex value of Wigner-D-function 
    """       
    # Doicu, Light Scattering by Systems of Particles, p. 271ff (B.33ff)   
    if m >= 0 and m_prime >= 0:
        delta_m_mprime = 1
    elif m >= 0 and m_prime < 0:
        delta_m_mprime = (-1) ** m_prime
    elif m < 0 and m_prime >= 0:
        delta_m_mprime = (-1) ** m
    elif m < 0 and m_prime < 0:
        delta_m_mprime = (-1) ** (m + m_prime)
        
    wig_D = ((-1) ** (m + m_prime) * np.exp(1j * m * alpha) * delta_m_mprime * wigner_d(l, m, m_prime, beta, wdsympy) 
            * np.exp(1j * m_prime * gamma))

    # Mishchenko, Scattering, Absorption and Emission of Light by small Particles, p.367 (B.38)
#    wig_D = np.exp(-1j * m * alpha) * wigner_d(l, m, m_prime, beta) * np.exp(-1j * m_prime * gamma)    
   
    return wig_D


def rotation_matrix(alpha=None, beta=None, gamma=None, euler_angles=None):
    if euler_angles is not None:
        alpha = euler_angles[0]
        beta = euler_angles[1]
        gamma = euler_angles[2]
    rotation_matrix_3 = [[np.cos(gamma), np.sin(gamma), 0], [- np.sin(gamma), np.cos(gamma), 0], [0, 0, 1]]
    rotation_matrix_2 = [[np.cos(beta), 0, - np.sin(beta)], [0, 1, 0], [np.sin(beta), 0, np.cos(beta)]]
    rotation_matrix_1 = [[np.cos(alpha), np.sin(alpha), 0], [- np.sin(alpha), np.cos(alpha), 0], [0, 0, 1]] 
    return np.dot(rotation_matrix_3, np.dot(rotation_matrix_2, rotation_matrix_1))


def vector_rotation(r, alpha=None, beta=None, gamma=None, euler_angles=None):
    return np.dot(rotation_matrix(alpha, beta, gamma, euler_angles), r)


def inverse_vector_rotation(r, alpha=None, beta=None, gamma=None, euler_angles=None):
    return np.dot(np.linalg.inv(rotation_matrix(alpha, beta, gamma, euler_angles)), r)


"""This package contains classes and functions to represent the system of
linear equations that needs to be solved in order to solve the scattering
problem, see section 3.7 of [Egel 2018 dissertation].

Symbolically, the linear system can be written like

.. math::
    (1 - TW)b = Ta,

where :math:`T` is the transition matrices of the particles, :math:`W` is the
particle coupling matrix, :math:`b` are the (unknown) coefficients of the
scattered field in terms of an outgoing spherical wave expansion and :math:`a`
are the coefficients of the initial field in terms of a regular spherical wave
expansion."""

import sys
import time
import inspect
import numpy as np
import scipy.special
import scipy.sparse.linalg
import scipy.linalg
import scipy.interpolate
from tqdm import tqdm
import warnings
import smuthi.fields as flds
import smuthi.linearsystem.tmatrix.t_matrix as tmt
import smuthi.fields.expansions as fldex
import smuthi.utility.cuda as cu
import smuthi.linearsystem.particlecoupling.direct_coupling as dircoup
import smuthi.linearsystem.particlecoupling.layer_mediated_coupling as laycoup
import smuthi.linearsystem.particlecoupling.prepare_lookup as look
import smuthi.linearsystem.linear_system_cuda as cusrc
from numba import config, set_num_threads
import smuthi.periodicboundaries as pb
import smuthi.periodicboundaries.coupling_helper as pbch


iter_num = 0


class LinearSystem:
    """Manage the assembly and solution of the linear system of equations.

    Args:
        particle_list (list):   List of smuthi.particles.Particle objects
        initial_field (smuthi.initial_field.InitialField):   Initial field object
        layer_system (smuthi.layers.LayerSystem):   Stratified medium
        k_parallel (numpy.ndarray or str): in-plane wavenumber.
                                           If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        solver_type (str):  What solver to use? Options: 'LU' for LU factorization, 'gmres' for GMRES iterative solver
        store_coupling_matrix (bool):   If True (default), the coupling matrix is stored. Otherwise it is recomputed on
                                        the fly during each iteration of the solver.
        coupling_matrix_lookup_resolution (float or None): If type float, compute particle coupling by interpolation of
                                                           a lookup table with that spacial resolution. A smaller number
                                                           implies higher accuracy and memory footprint.
                                                           If None (default), don't use a lookup table but compute the
                                                           coupling directly. This is more suitable for a small particle
                                                           number.
        interpolator_kind (str): interpolation order to be used, e.g. 'linear' or 'cubic'. This argument is ignored if
                                 coupling_matrix_lookup_resolution is None. In general, cubic interpolation is more
                                 accurate but a bit slower than linear.
        periodicity (tuple):                 tuple (a1, a2) specifying two 3-dimensional lattice vectors in Carthesian coordinates
                                             with a1, a2 (numpy.ndarrays)
        ewald_sum_separation_parameter (float):     Used to separate the real and reciprocal lattice sums to evaluate
                                                    particle coupling in periodic lattices.
        number_of_threads_periodic (int or str):    sets the number of threats used in a simulation with periodic particle arrangements
                                                    if 'default', all available CPU cores are used 
                                                    if negative, all but number_of_threads_periodic are used
        use_pvwf_coupling (bool):              If set to True, plane wave coupling is used to calculate
                                               the direct. Currently only possible in combination with direct solver strategy.
        pvwf_coupling_k_parallel (array):      k-parallel for PVWF coupling
    """

    def __init__(self,
                 particle_list,
                 initial_field,
                 layer_system,
                 k_parallel='default',
                 solver_type='LU',
                 solver_tolerance=1e-4,
                 store_coupling_matrix=True,
                 coupling_matrix_lookup_resolution=None,
                 interpolator_kind='cubic',
                 cuda_blocksize=None,
                 periodicity=None,
                 ewald_sum_separation_parameter='default',
                 number_of_threads_periodic='default',
                 use_pvwf_coupling=False,
                 pvwf_coupling_k_parallel=None):

        if cuda_blocksize is None:
            cuda_blocksize = cu.default_blocksize

        self.particle_list = particle_list
        self.initial_field = initial_field
        self.layer_system = layer_system
        self.k_parallel = k_parallel
        self.solver_type = solver_type
        self.solver_tolerance = solver_tolerance
        self.store_coupling_matrix = store_coupling_matrix
        self.coupling_matrix_lookup_resolution = coupling_matrix_lookup_resolution
        self.interpolator_kind = interpolator_kind
        self.cuda_blocksize = cuda_blocksize
        self.periodicity = periodicity
        self.ewald_sum_separation_parameter = ewald_sum_separation_parameter
        self.number_of_threads_periodic = number_of_threads_periodic
        self.use_pvwf_coupling = use_pvwf_coupling
        self.pvwf_coupling_k_parallel = pvwf_coupling_k_parallel

        if use_pvwf_coupling and (coupling_matrix_lookup_resolution is not None):
            warnings.warn(
                "The use_pvwf_coupling flag was defined in combination with a  "
                "lookup. This combination is not yet available. Ignoring the "
                "use_pvwf_coupling flag.")

        dummy_matrix = SystemMatrix(self.particle_list)
        sys.stdout.write('Number of unknowns: %i\n' % dummy_matrix.shape[0])

    def prepare(self):
        self.compute_initial_field_coefficients()
        self.compute_t_matrix()
        self.compute_coupling_matrix()
        self.master_matrix = MasterMatrix(t_matrix=self.t_matrix,
                                          coupling_matrix=self.coupling_matrix)

    def compute_initial_field_coefficients(self):
        """Evaluate initial field coefficients."""
        for particle in tqdm(self.particle_list,
                             desc='Initial field coefficients',
                             file=sys.stdout,
                             bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}'):
            particle.initial_field = self.initial_field.spherical_wave_expansion(particle, self.layer_system)

    def compute_t_matrix(self):
        """Initialize T-matrix object."""
        for particle in tqdm(self.particle_list,
                             desc='T-matrices                ',
                             file=sys.stdout,
                             bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}'):
            iS = self.layer_system.layer_number(particle.position[2])
            niS = self.layer_system.refractive_indices[iS]
            particle.t_matrix = particle.compute_t_matrix(self.initial_field.vacuum_wavelength, niS)
            if not particle.euler_angles == [0, 0, 0]:
                particle.t_matrix = tmt.rotate_t_matrix(particle.t_matrix, particle.l_max, particle.m_max,
                                                        particle.euler_angles, wdsympy=False)
        self.t_matrix = TMatrix(particle_list=self.particle_list)

    def compute_coupling_matrix(self):
        """Initialize coupling matrix object."""

        # LU factorization only possible with explicit matrix.
        if self.solver_type == 'LU':
            self.store_coupling_matrix = True

        if self.coupling_matrix_lookup_resolution is not None:
            z_list = [particle.position[2] for particle in self.particle_list]
            is_list = [self.layer_system.layer_number(z) for z in z_list]
            if not is_list.count(is_list[0]) == len(is_list):  # all particles in same layer?
                warnings.warn("Particles are not all in same layer. "
                              "Fall back to direct coupling matrix computation (no lookup).")
                self.coupling_matrix_lookup_resolution = None
            if self.periodicity is not None:
                warnings.warn("Periodic particle arrangement defined. "
                              "Disabling lookup.")
                self.coupling_matrix_lookup_resolution = None
            if self.store_coupling_matrix:
                warnings.warn("Explicit matrix computation using lookup currently not implemented. "
                              "Disabling lookup.")
                self.coupling_matrix_lookup_resolution = None
            else:  # use lookup
                if not self.interpolator_kind in ('linear', 'cubic'):
                    warnings.warn(self.interpolator_kind + ' interpolation not implemented. '
                                                           'Use "linear" instead')
                    self.interpolator_kind = 'linear'

                z_list = [particle.position[2] for particle in self.particle_list]
                if z_list.count(z_list[0]) == len(z_list):  # all particles at same height: use radial lookup
                    if cu.use_gpu:
                        sys.stdout.write('Coupling matrix computation by ' + self.interpolator_kind
                                         + ' interpolation of radial lookup on GPU.\n')
                        sys.stdout.flush()
                        self.coupling_matrix = CouplingMatrixRadialLookupCUDA(
                            vacuum_wavelength=self.initial_field.vacuum_wavelength,
                            particle_list=self.particle_list,
                            layer_system=self.layer_system,
                            k_parallel=self.k_parallel,
                            resolution=self.coupling_matrix_lookup_resolution,
                            cuda_blocksize=self.cuda_blocksize,
                            interpolator_kind=self.interpolator_kind)
                    else:
                        sys.stdout.write('Coupling matrix computation by ' + self.interpolator_kind
                                         + ' interpolation of radial lookup on CPU.\n')
                        sys.stdout.flush()
                        self.coupling_matrix = CouplingMatrixRadialLookupCPU(
                            vacuum_wavelength=self.initial_field.vacuum_wavelength,
                            particle_list=self.particle_list,
                            layer_system=self.layer_system,
                            k_parallel=self.k_parallel,
                            resolution=self.coupling_matrix_lookup_resolution,
                            interpolator_kind=self.interpolator_kind)
                else:  # not all particles at same height: use volume lookup
                    if cu.use_gpu:
                        sys.stdout.write('Coupling matrix computation by ' + self.interpolator_kind
                                         + ' interpolation of 3D lookup on GPU.\n')
                        sys.stdout.flush()
                        self.coupling_matrix = CouplingMatrixVolumeLookupCUDA(
                            vacuum_wavelength=self.initial_field.vacuum_wavelength,
                            particle_list=self.particle_list,
                            layer_system=self.layer_system,
                            k_parallel=self.k_parallel,
                            resolution=self.coupling_matrix_lookup_resolution,
                            interpolator_kind=self.interpolator_kind)
                    else:
                        sys.stdout.write('Coupling matrix computation by ' + self.interpolator_kind
                                         + ' interpolation of 3D lookup on CPU.\n')

                        sys.stdout.flush()
                        self.coupling_matrix = CouplingMatrixVolumeLookupCPU(
                            vacuum_wavelength=self.initial_field.vacuum_wavelength,
                            particle_list=self.particle_list,
                            layer_system=self.layer_system,
                            k_parallel=self.k_parallel,
                            resolution=self.coupling_matrix_lookup_resolution,
                            interpolator_kind=self.interpolator_kind)

        if self.coupling_matrix_lookup_resolution is None:
            if not self.store_coupling_matrix:
                warnings.warn("With lookup disabled, coupling matrix needs to be stored.")
                self.store_coupling_matrix = True
                
            if self.periodicity:
                sys.stdout.write('Explicit, periodic coupling matrix computation on CPU.\n')
                sys.stdout.flush()
                self.coupling_matrix = CouplingMatrixPeriodicGridNumba(
                    initial_field=self.initial_field,
                    particle_list=self.particle_list,
                    layer_system=self.layer_system,
                    periodicity=self.periodicity,
                    ewald_sum_separation_parameter=self.ewald_sum_separation_parameter,
                    num_threads=self.number_of_threads_periodic)
            
            else:
                sys.stdout.write('Explicit coupling matrix computation on CPU.\n')
                sys.stdout.flush()
                self.coupling_matrix = CouplingMatrixExplicit(
                    vacuum_wavelength=self.initial_field.vacuum_wavelength,
                    particle_list=self.particle_list,
                    layer_system=self.layer_system,
                    k_parallel=self.k_parallel,
                    use_pvwf_coupling=self.use_pvwf_coupling,
                    pvwf_coupling_k_parallel=self.pvwf_coupling_k_parallel)

    def solve(self):
        """Compute scattered field coefficients and store them
        in the particles' spherical wave expansion objects."""
        if len(self.particle_list) > 0:
            if self.solver_type.lower() == 'lu':
                sys.stdout.write('Solve (LU decomposition)  : ...')
                if not hasattr(self.master_matrix.linear_operator, 'A'):
                    raise ValueError('LU factorization only possible '
                                     'with the option "store coupling matrix".')
                if not hasattr(self.master_matrix, 'LU_piv'):
                    lu, piv = scipy.linalg.lu_factor(self.master_matrix.linear_operator.A,
                                                     overwrite_a=False)
                    self.master_matrix.LU_piv = (lu, piv)
                b = scipy.linalg.lu_solve(self.master_matrix.LU_piv,
                                          self.t_matrix.right_hand_side())
                sys.stdout.write(' done\n')
                sys.stdout.flush()
            elif self.solver_type.lower() == 'gmres':
                rhs = self.t_matrix.right_hand_side()
                start_time = time.time()
                iter_num = 0

                def status_msg(rk):
                    nonlocal iter_num
                    iter_msg = ('Solve (GMRES)             : Iter ' + str(iter_num)
                                + ' | Rel. residual: '
                                + "{:.2e}".format(np.linalg.norm(rk))
                                + ' | elapsed: ' + str(int(time.time() - start_time)) + 's')
                    sys.stdout.write('\r' + iter_msg)
                    iter_num += 1

                b, info = scipy.sparse.linalg.gmres(self.master_matrix.linear_operator, rhs, rhs,
                                                    tol=self.solver_tolerance, callback=status_msg)
            #                sys.stdout.write('\n')
            elif self.solver_type.lower() == 'lgmres':
                rhs = self.t_matrix.right_hand_side()
                start_time = time.time()
                iter_num = 0

                def status_msg(xk):
                    frame = inspect.currentframe().f_back
                    # for some reason, lgmres passes xk to callback instead of rk ...
                    # then, we could inspect r_norm from the currentframe, but it's not there ... (?)
                    # therefore we have to re-calculate it again, which is a total waste of time ...
                    rk = frame.f_locals['nrm2'](frame.f_locals['r_outer'])/frame.f_locals['b_norm']
                    nonlocal iter_num
                    iter_msg = ('Solve (LGMRES)            : Iter ' + str(iter_num)
                                + ' | Rel. residual: '
                                + "{:.2e}".format(np.linalg.norm(rk))
                                + ' | elapsed: ' + str(int(time.time() - start_time)) + 's')
                    sys.stdout.write('\r' + iter_msg)
                    iter_num += 1

                b, info = scipy.sparse.linalg.lgmres(self.master_matrix.linear_operator, rhs, rhs,
                                                    tol=self.solver_tolerance, callback=status_msg)
            #                sys.stdout.write('\n')
            elif self.solver_type.lower() == 'gcrotmk':
                rhs = self.t_matrix.right_hand_side()
                start_time = time.time()
                iter_num = 0

                def status_msg(xk):
                    frame = inspect.currentframe().f_back
                    # also gcrotmk passes xk to callback instead of rk ...
                    # let's calculate residuals on the fly also in this case ...
                    rk = frame.f_locals['nrm2'](frame.f_locals['r'])/frame.f_locals['b_norm']
                    nonlocal iter_num
                    iter_msg = ('Solve (GCROTMK)           : Iter ' + str(iter_num)
                                + ' | Rel. residual: '
                                + "{:.2e}".format(np.linalg.norm(rk))
                                + ' | elapsed: ' + str(int(time.time() - start_time)) + 's')
                    sys.stdout.write('\r' + iter_msg)
                    iter_num += 1

                b, info = scipy.sparse.linalg.gcrotmk(self.master_matrix.linear_operator, rhs, rhs,
                                                    tol=self.solver_tolerance, callback=status_msg)
            #                sys.stdout.write('\n')
            else:
                raise ValueError('This solver type is currently not implemented.')

        for iS, particle in enumerate(self.particle_list):
            i_iS = self.layer_system.layer_number(particle.position[2])
            n_iS = self.layer_system.refractive_indices[i_iS]
            k = flds.angular_frequency(self.initial_field.vacuum_wavelength) * n_iS
            loz, upz = self.layer_system.lower_zlimit(i_iS), self.layer_system.upper_zlimit(i_iS)
            particle.scattered_field = fldex.SphericalWaveExpansion(k=k, l_max=particle.l_max, m_max=particle.m_max,
                                                                    kind='outgoing', reference_point=particle.position,
                                                                    lower_z=loz, upper_z=upz)
            particle.scattered_field.coefficients = b[self.master_matrix.index_block(iS)]
            particle.scattered_field.validity_conditions.append(particle.is_outside)
            self.initial_field.validity_conditions.append(particle.is_outside)


class SystemMatrix:
    r"""A system matrix is an abstract linear operator that operates on a
    system coefficient vector, i.e. a vector :math:`c = c_{\tau,l,m}^i`, where
    :math:`(\tau, l, m)` are the multipole indices and :math:`i` indicates the
    particle number. In other words, if we have a spherical wave expansion for
    each particle, and write all the expansion coefficients of these expansions
    into one (long) array, what we get is a system vector.
    """

    def __init__(self, particle_list):
        self.particle_list = particle_list
        blocksizes = [flds.blocksize(particle.l_max, particle.m_max) for particle in self.particle_list]
        self.shape = (sum(blocksizes), sum(blocksizes))

    def index_block(self, i):
        """
        Args:
            i (int): number of particle

        Returns:
            indices that correspond to the coefficients for that particle
        """
        blocksizes = [flds.blocksize(particle.l_max, particle.m_max) for particle in self.particle_list[:(i + 1)]]
        return range(sum(blocksizes[:i]), sum(blocksizes))

    def index(self, i, tau, l, m):
        r"""
        Args:
            i (int):    particle number
            tau (int):    spherical polarization index
            l (int):    multipole degree
            m (int):    multipole order

        Returns:
            Position in a system vector that corresponds to the :math:`(\tau, l, m)` coefficient of the i-th particle.
        """
        blocksizes = [flds.blocksize(particle.l_max, particle.m_max) for particle in self.particle_list[:i]]
        return sum(blocksizes) + flds.multi_to_single_index(tau, l, m, self.particle_list[i].l_max,
                                                             self.particle_list[i].m_max)


class MasterMatrix(SystemMatrix):
    r"""Represent the master matrix :math:`M = 1 - TW` as a linear operator.

    Args:
        t_matrix (SystemMatrix):          System T-matrix
        coupling_matrix (SystemMatrix):   System coupling matrix
    """

    def __init__(self, t_matrix, coupling_matrix):
        SystemMatrix.__init__(self, t_matrix.particle_list)
        if type(coupling_matrix.linear_operator).__name__ == 'MatrixLinearOperator':
            M = (np.eye(self.shape[0], dtype=complex)
                 - t_matrix.linear_operator.matmat(coupling_matrix.linear_operator.A))
            self.linear_operator = scipy.sparse.linalg.aslinearoperator(M)
        else:
            def apply_master_matrix(vector):
                return vector - t_matrix.linear_operator.dot(coupling_matrix.linear_operator.matvec(vector))

            self.linear_operator = scipy.sparse.linalg.LinearOperator(shape=self.shape, matvec=apply_master_matrix,
                                                                      dtype=complex)


class TMatrix(SystemMatrix):
    """Collect the particle T-matrices in a global lienear operator.

    Args:
        particle_list (list):   List of smuthi.particles.Particle objects containing a t_matrix attribute.
    """

    def __init__(self, particle_list):
        SystemMatrix.__init__(self, particle_list)

        def apply_t_matrix(vector):
            tv = np.zeros(vector.shape, dtype=complex)
            for i_s, particle in enumerate(particle_list):
                tv[self.index_block(i_s)] = particle.t_matrix.dot(vector[self.index_block(i_s)])
            return tv

        self.linear_operator = scipy.sparse.linalg.LinearOperator(shape=self.shape, matvec=apply_t_matrix,
                                                                  matmat=apply_t_matrix, dtype=complex)

    def right_hand_side(self):
        r"""The right hand side of the linear system is given by :math:`\sum_{\tau l m} T^i_{\tau l m} a^i_{\tau l m }`

        Returns:
            right hand side as a complex numpy.ndarray
        """
        tai = np.zeros(self.shape[0], dtype=complex)
        for i_s, particle in enumerate(self.particle_list):
            tai[self.index_block(i_s)] = particle.t_matrix.dot(particle.initial_field.coefficients)
        return tai


class CouplingMatrixExplicit(SystemMatrix):
    """Class for an explicit representation of the coupling matrix. Recommended for small particle numbers.

    Args:
        vacuum_wavelength (float):  Vacuum wavelength in length units
        particle_list (list):   List of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):   Stratified medium
        k_parallell (numpy.ndarray or str): In-plane wavenumber.
                                            If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
    """

    def __init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel='default', use_pvwf_coupling=False, pvwf_coupling_k_parallel=None):

        SystemMatrix.__init__(self, particle_list)
        coup_mat = np.zeros(self.shape, dtype=complex)
        sys.stdout.write('Coupling matrix memory footprint: ' + look.size_format(coup_mat.nbytes) + '\n')
        sys.stdout.flush()
        
        if layer_system.is_degenerate():
            sys.stdout.write('Omitting calculations of layer mediated coupling since no difference in layers was detected. \n')
            sys.stdout.flush()
        
        for s1, particle1 in enumerate(tqdm(particle_list, desc='Particle coupling matrix  ', file=sys.stdout,
                                            bar_format='{l_bar}{bar}| elapsed: {elapsed} ' 'remaining: {remaining}')):
            idx1 = np.array(self.index_block(s1))[:, None]
            for s2, particle2 in enumerate(particle_list):
                idx2 = self.index_block(s2)

                # direct contribution
                if use_pvwf_coupling:
                    coup_mat[idx1, idx2] = dircoup.direct_coupling_block_pvwf_mediated(vacuum_wavelength,
                                                                                       particle1,
                                                                                       particle2,
                                                                                       layer_system,
                                                                                       pvwf_coupling_k_parallel)
                else:
                    coup_mat[idx1, idx2] = dircoup.direct_coupling_block(vacuum_wavelength,
                                                                         particle1,
                                                                         particle2,
                                                                         layer_system)

                # layer mediated contribution
                if not layer_system.is_degenerate():
                    coup_mat[idx1, idx2] += laycoup.layer_mediated_coupling_block(vacuum_wavelength, particle1, particle2,
                                                                                  layer_system, k_parallel)
        self.linear_operator = scipy.sparse.linalg.aslinearoperator(coup_mat)


class CouplingMatrixPeriodicGridNumba(SystemMatrix):
    """ Class for an explicit representation of the coupling matrix of periodic particle arrangements.
        Computation supports Numba.
    Args:
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        particle_list (list):                               list of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):           stratified medium
        periodicity (tuple):                                (a1, a2) lattice vector 1 and 2 in carthesian coordinates
        ewald_sum_separation_parameter (float):             Ewald sum separation parameter
        num_threads (int or str):                           if 'default' all available CPU cores are used
                                                            if negative, all but num_threads are used
    """
    
    def __init__(self, initial_field, particle_list, layer_system, periodicity,
                 ewald_sum_separation_parameter, num_threads='default'):
        
        max_num_threads = config.NUMBA_DEFAULT_NUM_THREADS
        if type(num_threads).__name__ == 'int':
            if 0 < num_threads <= max_num_threads:
                set_num_threads(num_threads)
                sys.stdout.write('Parallel Numba execution limited to %d of %d available CPU cores. \n' % (num_threads, max_num_threads))
            elif num_threads <= 0:
                set_num_threads(max_num_threads + num_threads)
                sys.stdout.write('Parallel Numba execution limited to %d of %d available CPU cores. \n' % (max_num_threads + num_threads, max_num_threads))
            else:
                sys.stdout.write('Parallel Numba execution uses all %d available CPU cores. \n' % (max_num_threads))
        else:
            sys.stdout.write('Parallel Numba execution uses all %d available CPU cores. \n' % (max_num_threads))
        sys.stdout.flush()
        
        a1, a2 = periodicity
        
        if ewald_sum_separation_parameter == 'default':
            eta = pb.default_Ewald_sum_separation
        else:
            eta = ewald_sum_separation_parameter
        # prepare a5b5 lookup table
        lmax_global = np.max([particle.l_max for particle in particle_list])
        mmax_global = np.max([particle.m_max for particle in particle_list])
        a5b5_mat = pbch.a5b5_lookup(lmax_global, lmax_global, mmax_global, mmax_global)
        
        i_sca = layer_system.layer_number(particle_list[0].position[2]) # all particles are within one layer
        if initial_field.polar_angle < np.pi:
            pwe_exc = initial_field.plane_wave_expansion(layer_system, i_sca)[0]
        else:
            pwe_exc = initial_field.plane_wave_expansion(layer_system, i_sca)[1]
        k0t = np.array([pwe_exc.k_parallel[0] * np.cos(pwe_exc.azimuthal_angles)[0],
                        pwe_exc.k_parallel[0] * np.sin(pwe_exc.azimuthal_angles)[0]]) 
        
        num_particles = len(particle_list)
        positions = np.zeros([num_particles, 3], np.float64)
        radii = np.zeros([num_particles], np.float64)
        lmax_array = np.zeros([num_particles], np.int32)
        mmax_array = np.zeros([num_particles], np.int32)
        for idx, particle in enumerate(particle_list):
            positions[idx] = particle.position
            lmax_array[idx] = particle.l_max
            mmax_array[idx] = particle.m_max
            if type(particle).__name__ == 'Sphere':
                radii[idx] = particle.radius
            elif type(particle).__name__ == 'FiniteCylinder':
                radii[idx] = np.sqrt(particle.cylinder_radius ** 2 + (particle.cylinder_height / 2) ** 2)

        import smuthi.periodicboundaries.particle_coupling as pbcoup
        coup_mat = pbcoup.periodic_coupling_matrix(initial_field.vacuum_wavelength, k0t,
                            pwe_exc.azimuthal_angles[0], np.array(layer_system.thicknesses, np.float64),
                            np.array(layer_system.refractive_indices, np.complex128),
                            i_sca, positions, radii, lmax_array, mmax_array, a1, a2, eta, a5b5_mat, mmax_global)
        sys.stdout.write('Coupling matrix computation finished. \n')
        sys.stdout.flush()
        
        self.linear_operator = scipy.sparse.linalg.aslinearoperator(coup_mat)


class CouplingMatrixVolumeLookup(SystemMatrix):
    """Base class for 3D lookup based coupling matrix either on CPU or on GPU (CUDA).

    Args:
        vacuum_wavelength (float): vacuum wavelength in length units
        particle_list (list): list of sumthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem): stratified medium
        k_parallel (numpy.ndarray or str): in-plane wavenumber.
                                           If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float or None): spatial resolution of the lookup in the radial direction
    """

    def __init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel='default', resolution=None):
        z_list = [particle.position[2] for particle in particle_list]
        is_list = [layer_system.layer_number(z) for z in z_list]
        assert is_list.count(is_list[0]) == len(is_list)  # all particles in same layer?

        SystemMatrix.__init__(self, particle_list)

        self.l_max = max([particle.l_max for particle in particle_list])
        self.m_max = max([particle.m_max for particle in particle_list])
        self.blocksize = flds.blocksize(self.l_max, self.m_max)
        self.resolution = resolution
        lkup = look.volumetric_coupling_lookup_table(vacuum_wavelength=vacuum_wavelength, particle_list=particle_list,
                                                     layer_system=layer_system, k_parallel=k_parallel,
                                                     resolution=resolution)
        self.lookup_table_plus, self.lookup_table_minus = lkup[0], lkup[1]
        self.rho_array, self.sum_z_array, self.diff_z_array = lkup[2], lkup[3], lkup[4]


class CouplingMatrixVolumeLookupCPU(CouplingMatrixVolumeLookup):
    """Class for 3D lookup based coupling matrix running on CPU. This is used when no suitable GPU device is detected
    or when PyCuda is not installed.

    Args:
        vacuum_wavelength (float): vacuum wavelength in length units
        particle_list (list): list of sumthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem): stratified medium
        k_parallel (numpy.ndarray or str): in-plane wavenumber.
                                           If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float or None): spatial resolution of the lookup in the radial direction
        interpolator_kind (str): 'linear' or 'cubic' interpolation
    """

    def __init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel='default', resolution=None,
                 interpolator_kind='cubic'):

        if interpolator_kind == 'cubic':
            interpolation_order = 3
        else:
            interpolation_order = 1

        CouplingMatrixVolumeLookup.__init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel,
                                            resolution)

        x_array = np.array([particle.position[0] for particle in particle_list])
        y_array = np.array([particle.position[1] for particle in particle_list])
        z_array = np.array([particle.position[2] for particle in particle_list])

        self.particle_rho_array = np.sqrt((x_array[:, None] - x_array[None, :]) ** 2
                                          + (y_array[:, None] - y_array[None, :]) ** 2)
        self.particle_phi_array = np.arctan2(y_array[:, None] - y_array[None, :], x_array[:, None] - x_array[None, :])
        self.particle_sz_array = z_array[:, None] + z_array[None, :]
        self.particle_dz_array = z_array[:, None] - z_array[None, :]

        # contains for each n all positions in the large system arrays that correspond to n:
        self.system_vector_index_list = [[] for i in range(self.blocksize)]

        # same size as system_vector_index_list, contains the according particle numbers:
        self.particle_number_list = [[] for i in range(self.blocksize)]
        self.m_list = [None for i in range(self.blocksize)]
        for i, particle in enumerate(particle_list):
            for m in range(-particle.m_max, particle.m_max + 1):
                for l in range(max(1, abs(m)), particle.l_max + 1):
                    for tau in range(2):
                        n_lookup = flds.multi_to_single_index(tau=tau, l=l, m=m, l_max=self.l_max, m_max=self.m_max)
                        self.system_vector_index_list[n_lookup].append(self.index(i, tau, l, m))
                        self.particle_number_list[n_lookup].append(i)
                        self.m_list[n_lookup] = m
        for n in range(self.blocksize):
            self.system_vector_index_list[n] = np.array(self.system_vector_index_list[n])
            self.particle_number_list[n] = np.array(self.particle_number_list[n])

        self.lookup_plus_real = [[None for i in range(self.blocksize)] for i2 in range(self.blocksize)]
        self.lookup_plus_imag = [[None for i in range(self.blocksize)] for i2 in range(self.blocksize)]
        self.lookup_minus_real = [[None for i in range(self.blocksize)] for i2 in range(self.blocksize)]
        self.lookup_minus_imag = [[None for i in range(self.blocksize)] for i2 in range(self.blocksize)]
        for n1 in range(self.blocksize):
            for n2 in range(self.blocksize):
                self.lookup_plus_real[n1][n2] = scipy.interpolate.RectBivariateSpline(
                    x=self.rho_array, y=self.sum_z_array, z=self.lookup_table_plus[:, :, n1, n2].real,
                    kx=interpolation_order, ky=interpolation_order)
                self.lookup_plus_imag[n1][n2] = scipy.interpolate.RectBivariateSpline(
                    x=self.rho_array, y=self.sum_z_array, z=self.lookup_table_plus[:, :, n1, n2].imag,
                    kx=interpolation_order, ky=interpolation_order)
                self.lookup_minus_real[n1][n2] = scipy.interpolate.RectBivariateSpline(
                    x=self.rho_array, y=self.diff_z_array, z=self.lookup_table_minus[:, :, n1, n2].real,
                    kx=interpolation_order, ky=interpolation_order)
                self.lookup_minus_imag[n1][n2] = scipy.interpolate.RectBivariateSpline(
                    x=self.rho_array, y=self.diff_z_array, z=self.lookup_table_minus[:, :, n1, n2].imag,
                    kx=interpolation_order, ky=interpolation_order)

        def matvec(in_vec):
            out_vec = np.zeros(shape=in_vec.shape, dtype=complex)
            for n1 in range(self.blocksize):
                i1 = self.particle_number_list[n1]
                idx1 = self.system_vector_index_list[n1]
                m1 = self.m_list[n1]
                for n2 in range(self.blocksize):
                    i2 = self.particle_number_list[n2]
                    idx2 = self.system_vector_index_list[n2]
                    m2 = self.m_list[n2]
                    rho = self.particle_rho_array[i1[:, None], i2[None, :]]
                    phi = self.particle_phi_array[i1[:, None], i2[None, :]]
                    sz = self.particle_sz_array[i1[:, None], i2[None, :]]
                    dz = self.particle_dz_array[i1[:, None], i2[None, :]]
                    Mpl = self.lookup_plus_real[n1][n2].ev(rho, sz) + 1j * self.lookup_plus_imag[n1][n2].ev(rho, sz)
                    Mmn = self.lookup_minus_real[n1][n2].ev(rho, dz) + 1j * self.lookup_minus_imag[n1][n2].ev(rho, dz)
                    M = (Mpl + Mmn) * np.exp(1j * (m2 - m1) * phi)
                    out_vec[idx1] += M.dot(in_vec[idx2])
            return out_vec

        self.linear_operator = scipy.sparse.linalg.LinearOperator(shape=self.shape, matvec=matvec, dtype=complex)


class CouplingMatrixVolumeLookupCUDA(CouplingMatrixVolumeLookup):
    """Class for 3D lookup based coupling matrix running on GPU.

    Args:
        vacuum_wavelength (float): vacuum wavelength in length units
        particle_list (list): list of sumthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem): stratified medium
        k_parallel (numpy.ndarray or str): in-plane wavenumber.
                                           If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float or None): spatial resolution of the lookup in the radial direction
        cuda_blocksize (int): threads per block for cuda call
        interpolator_kind (str): 'linear' (default) or 'cubic' interpolation
    """

    def __init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel='default', resolution=None,
                 cuda_blocksize=None, interpolator_kind='linear'):

        if cuda_blocksize is None:
            cuda_blocksize = cu.default_blocksize

        CouplingMatrixVolumeLookup.__init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel,
                                            resolution)

        sys.stdout.write('Prepare CUDA kernel and device lookup data ... ')
        sys.stdout.flush()
        start_time = time.time()

        if interpolator_kind == 'linear':
            coupling_source = cusrc.linear_volume_lookup_source % (self.blocksize, self.shape[0], len(self.sum_z_array),
                                                                min(self.rho_array), min(self.sum_z_array),
                                                                min(self.diff_z_array), self.resolution)
        elif interpolator_kind == 'cubic':
            coupling_source = cusrc.cubic_volume_lookup_source % (self.blocksize, self.shape[0], len(self.sum_z_array),
                                                               min(self.rho_array), min(self.sum_z_array),
                                                               min(self.diff_z_array), self.resolution)

        coupling_function = cu.SourceModule(coupling_source).get_function("coupling_kernel")

        n_lookup_array = np.zeros(self.shape[0], dtype=np.uint32)
        m_particle_array = np.zeros(self.shape[0], dtype=np.float32)
        x_array = np.zeros(self.shape[0], dtype=np.float32)
        y_array = np.zeros(self.shape[0], dtype=np.float32)
        z_array = np.zeros(self.shape[0], dtype=np.float32)

        i_particle = 0
        for i, particle in enumerate(particle_list):
            for m in range(-particle.m_max, particle.m_max + 1):
                for l in range(max(1, abs(m)), particle.l_max + 1):
                    for tau in range(2):
                        i_taulm = flds.multi_to_single_index(tau, l, m, particle.l_max, particle.m_max)
                        idx = i_particle + i_taulm

                        n_lookup_array[idx] = flds.multi_to_single_index(tau, l, m, self.l_max, self.m_max)
                        m_particle_array[idx] = m

                        # scale the x and y position to the lookup resolution:
                        x_array[idx] = particle.position[0]
                        y_array[idx] = particle.position[1]
                        z_array[idx] = particle.position[2]

            i_particle += flds.blocksize(particle.l_max, particle.m_max)

        re_lookup_pl = self.lookup_table_plus.real.astype(dtype=np.float32)
        im_lookup_pl = self.lookup_table_plus.imag.astype(dtype=np.float32)
        re_lookup_mn = self.lookup_table_minus.real.astype(dtype=np.float32)
        im_lookup_mn = self.lookup_table_minus.imag.astype(dtype=np.float32)

        # transfer data to gpu
        n_lookup_array_d = cu.gpuarray.to_gpu(n_lookup_array)
        m_particle_array_d = cu.gpuarray.to_gpu(m_particle_array)
        x_array_d = cu.gpuarray.to_gpu(x_array)
        y_array_d = cu.gpuarray.to_gpu(y_array)
        z_array_d = cu.gpuarray.to_gpu(z_array)
        re_lookup_pl_d = cu.gpuarray.to_gpu(re_lookup_pl)
        im_lookup_pl_d = cu.gpuarray.to_gpu(im_lookup_pl)
        re_lookup_mn_d = cu.gpuarray.to_gpu(re_lookup_mn)
        im_lookup_mn_d = cu.gpuarray.to_gpu(im_lookup_mn)

        sys.stdout.write('done | elapsed: ' + str(int(time.time() - start_time)) + 's\n')
        sys.stdout.flush()

        cuda_gridsize = (self.shape[0] + cuda_blocksize - 1) // cuda_blocksize

        def matvec(in_vec):
            re_in_vec_d = cu.gpuarray.to_gpu(np.float32(in_vec.real))
            im_in_vec_d = cu.gpuarray.to_gpu(np.float32(in_vec.imag))
            re_result_d = cu.gpuarray.zeros(in_vec.shape, dtype=np.float32)
            im_result_d = cu.gpuarray.zeros(in_vec.shape, dtype=np.float32)
            coupling_function(n_lookup_array_d.gpudata, m_particle_array_d.gpudata, x_array_d.gpudata,
                              y_array_d.gpudata, z_array_d.gpudata, re_lookup_pl_d.gpudata, im_lookup_pl_d.gpudata,
                              re_lookup_mn_d.gpudata, im_lookup_mn_d.gpudata, re_in_vec_d.gpudata, im_in_vec_d.gpudata,
                              re_result_d.gpudata, im_result_d.gpudata, block=(cuda_blocksize, 1, 1),
                              grid=(cuda_gridsize, 1))
            return re_result_d.get() + 1j * im_result_d.get()

        self.linear_operator = scipy.sparse.linalg.LinearOperator(shape=self.shape, matvec=matvec, dtype=complex)


class CouplingMatrixRadialLookup(SystemMatrix):
    """Base class for radial lookup based coupling matrix either on CPU or on GPU (CUDA).

    Args:
        vacuum_wavelength (float): vacuum wavelength in length units
        particle_list (list): list of sumthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem): stratified medium
        k_parallel (numpy.ndarray or str): in-plane wavenumber.
                                           If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float or None): spatial resolution of the lookup in the radial direction
    """

    def __init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel='default', resolution=None):
        z_list = [particle.position[2] for particle in particle_list]
        assert z_list.count(z_list[0]) == len(z_list)

        SystemMatrix.__init__(self, particle_list)

        self.l_max = max([particle.l_max for particle in particle_list])
        self.m_max = max([particle.m_max for particle in particle_list])
        self.blocksize = flds.blocksize(self.l_max, self.m_max)
        self.resolution = resolution
        self.lookup_table, self.radial_distance_array = look.radial_coupling_lookup_table(
            vacuum_wavelength=vacuum_wavelength, particle_list=particle_list, layer_system=layer_system,
            k_parallel=k_parallel, resolution=resolution)


class CouplingMatrixRadialLookupCUDA(CouplingMatrixRadialLookup):
    """Radial lookup based coupling matrix either on GPU (CUDA).

    Args:
        vacuum_wavelength (float): vacuum wavelength in length units
        particle_list (list): list of sumthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem): stratified medium
        k_parallel (numpy.ndarray or str): in-plane wavenumber.
                                           If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float or None): spatial resolution of the lookup in the radial direction
        cuda_blocksize (int): threads per block when calling CUDA kernel
    """

    def __init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel='default', resolution=None,
                 cuda_blocksize=None, interpolator_kind='linear'):

        if cuda_blocksize is None:
            cuda_blocksize = cu.default_blocksize

        CouplingMatrixRadialLookup.__init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel,
                                            resolution)

        sys.stdout.write('Prepare CUDA kernel and device lookup data ... ')
        sys.stdout.flush()
        start_time = time.time()

        if interpolator_kind == 'linear':
            coupling_source = cusrc.linear_radial_lookup_source % (self.blocksize, self.shape[0],
                                                                self.radial_distance_array.min(), resolution)
        elif interpolator_kind == 'cubic':
            coupling_source = cusrc.cubic_radial_lookup_source % (self.blocksize, self.shape[0],
                                                               self.radial_distance_array.min(), resolution)

        coupling_function = cu.SourceModule(coupling_source).get_function("coupling_kernel")

        n_lookup_array = np.zeros(self.shape[0], dtype=np.uint32)
        m_particle_array = np.zeros(self.shape[0], dtype=np.float32)
        x_array = np.zeros(self.shape[0], dtype=np.float32)
        y_array = np.zeros(self.shape[0], dtype=np.float32)

        i_particle = 0
        for i, particle in enumerate(particle_list):
            for m in range(-particle.m_max, particle.m_max + 1):
                for l in range(max(1, abs(m)), particle.l_max + 1):
                    for tau in range(2):
                        # idx = self.index(i, tau, l, m)
                        i_taulm = flds.multi_to_single_index(tau, l, m, particle.l_max, particle.m_max)
                        idx = i_particle + i_taulm

                        n_lookup_array[idx] = flds.multi_to_single_index(tau, l, m, self.l_max, self.m_max)
                        m_particle_array[idx] = m

                        # scale the x and y position to the lookup resolution:
                        x_array[idx] = particle.position[0]
                        y_array[idx] = particle.position[1]

            i_particle += flds.blocksize(particle.l_max, particle.m_max)

        # lookup as numpy array in required shape
        re_lookup = self.lookup_table.real.astype(np.float32)
        im_lookup = self.lookup_table.imag.astype(np.float32)

        # transfer data to gpu
        n_lookup_array_d = cu.gpuarray.to_gpu(n_lookup_array)
        m_particle_array_d = cu.gpuarray.to_gpu(m_particle_array)
        x_array_d = cu.gpuarray.to_gpu(x_array)
        y_array_d = cu.gpuarray.to_gpu(y_array)
        re_lookup_d = cu.gpuarray.to_gpu(re_lookup)
        im_lookup_d = cu.gpuarray.to_gpu(im_lookup)

        sys.stdout.write('done | elapsed: ' + str(int(time.time() - start_time)) + 's\n')
        sys.stdout.flush()

        cuda_gridsize = (self.shape[0] + cuda_blocksize - 1) // cuda_blocksize

        def matvec(in_vec):
            re_in_vec_d = cu.gpuarray.to_gpu(np.float32(in_vec.real))
            im_in_vec_d = cu.gpuarray.to_gpu(np.float32(in_vec.imag))
            re_result_d = cu.gpuarray.zeros(in_vec.shape, dtype=np.float32)
            im_result_d = cu.gpuarray.zeros(in_vec.shape, dtype=np.float32)
            coupling_function(n_lookup_array_d.gpudata, m_particle_array_d.gpudata, x_array_d.gpudata,
                              y_array_d.gpudata,
                              re_lookup_d.gpudata, im_lookup_d.gpudata, re_in_vec_d.gpudata, im_in_vec_d.gpudata,
                              re_result_d.gpudata, im_result_d.gpudata, block=(cuda_blocksize, 1, 1),
                              grid=(cuda_gridsize, 1))
            return re_result_d.get() + 1j * im_result_d.get()

        self.linear_operator = scipy.sparse.linalg.LinearOperator(shape=self.shape, matvec=matvec, dtype=complex)


class CouplingMatrixRadialLookupCPU(CouplingMatrixRadialLookup):
    """Class for radial lookup based coupling matrix running on CPU. This is used when no suitable GPU device is detected
    or when PyCuda is not installed.

    Args:
        vacuum_wavelength (float): vacuum wavelength in length units
        particle_list (list): list of sumthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem): stratified medium
        k_parallel (numpy.ndarray or str): in-plane wavenumber.
                                           If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float or None): spatial resolution of the lookup in the radial direction
        kind (str): interpolation order, e.g. 'linear' or 'cubic'
    """

    def __init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel='default', resolution=None,
                 interpolator_kind='linear'):

        z_list = [particle.position[2] for particle in particle_list]
        assert z_list.count(z_list[0]) == len(z_list)

        CouplingMatrixRadialLookup.__init__(self, vacuum_wavelength, particle_list, layer_system, k_parallel,
                                            resolution)

        x_array = np.array([particle.position[0] for particle in particle_list])
        y_array = np.array([particle.position[1] for particle in particle_list])

        self.particle_rho_array = np.sqrt((x_array[:, None] - x_array[None, :]) ** 2
                                          + (y_array[:, None] - y_array[None, :]) ** 2)
        self.particle_phi_array = np.arctan2(y_array[:, None] - y_array[None, :], x_array[:, None] - x_array[None, :])

        # contains for each n all positions in the large system arrays that correspond to n:
        self.system_vector_index_list = [[] for i in range(self.blocksize)]

        # same size as system_vector_index_list, contains the according particle numbers:
        self.particle_number_list = [[] for i in range(self.blocksize)]
        self.m_list = [None for i in range(self.blocksize)]
        for i, particle in enumerate(particle_list):
            for m in range(-particle.m_max, particle.m_max + 1):
                for l in range(max(1, abs(m)), particle.l_max + 1):
                    for tau in range(2):
                        n_lookup = flds.multi_to_single_index(tau=tau, l=l, m=m, l_max=self.l_max, m_max=self.m_max)
                        self.system_vector_index_list[n_lookup].append(self.index(i, tau, l, m))
                        self.particle_number_list[n_lookup].append(i)
                        self.m_list[n_lookup] = m
        for n in range(self.blocksize):
            self.system_vector_index_list[n] = np.array(self.system_vector_index_list[n])
            self.particle_number_list[n] = np.array(self.particle_number_list[n])

        lookup = [[None for i in range(self.blocksize)] for i2 in range(self.blocksize)]
        for n1 in range(self.blocksize):
            for n2 in range(self.blocksize):
                lookup[n1][n2] = scipy.interpolate.interp1d(x=self.radial_distance_array,
                                                            y=self.lookup_table[:, n1, n2],
                                                            kind=interpolator_kind, axis=-1, assume_sorted=True)

        def matvec(in_vec):
            out_vec = np.zeros(shape=in_vec.shape, dtype=complex)
            for n1 in range(self.blocksize):
                i1 = self.particle_number_list[n1]
                idx1 = self.system_vector_index_list[n1]
                m1 = self.m_list[n1]
                for n2 in range(self.blocksize):
                    i2 = self.particle_number_list[n2]
                    idx2 = self.system_vector_index_list[n2]
                    m2 = self.m_list[n2]
                    rho = self.particle_rho_array[i1[:, None], i2[None, :]]
                    phi = self.particle_phi_array[i1[:, None], i2[None, :]]
                    M = lookup[n1][n2](rho)
                    M = M * np.exp(1j * (m2 - m1) * phi)
                    out_vec[idx1] += M.dot(in_vec[idx2])
            return out_vec

        self.linear_operator = scipy.sparse.linalg.LinearOperator(shape=self.shape, matvec=matvec, dtype=complex)



# -*- coding: utf-8 -*-
import numpy as np
import smuthi.fields as flds
import smuthi.utility.math as sf
import smuthi.fields.transformations as trf


def mie_coefficient(tau, l, k_medium, k_particle, radius):
    """Return the Mie coefficients of a sphere.

    Args:
        tau         integer: spherical polarization, 0 for spherical TE and 1 for spherical TM
        l           integer: l=1,... multipole degree (polar quantum number)
        k_medium    float or complex: wavenumber in surrounding medium (inverse length unit)
        k_particle  float or complex: wavenumber inside sphere (inverse length unit)
        radius      float: radius of sphere (length unit)

    Returns:
        Mie coefficients as complex
    """
    jlkr_medium = sf.spherical_bessel(l, k_medium * radius)
    jlkr_particle = sf.spherical_bessel(l, k_particle * radius)
    dxxj_medium = sf.dx_xj(l, k_medium * radius)
    dxxj_particle = sf.dx_xj(l, k_particle * radius)

    hlkr_medium = sf.spherical_hankel(l, k_medium * radius)
    dxxh_medium = sf.dx_xh(l, k_medium * radius)

    if tau == 0:
        q = (jlkr_medium * dxxj_particle - jlkr_particle * dxxj_medium) / (jlkr_particle * dxxh_medium - hlkr_medium *
                                                                           dxxj_particle)
    elif tau == 1:
        q = ((k_medium ** 2 * jlkr_medium * dxxj_particle - k_particle ** 2 * jlkr_particle * dxxj_medium) /
             (k_particle ** 2 * jlkr_particle * dxxh_medium - k_medium ** 2 * hlkr_medium * dxxj_particle))
    else:
        raise ValueError('tau must be 0 (spherical TE) or 1 (spherical TM)')

    return q


def internal_mie_coefficient(tau, l, k_medium, k_particle, radius):
    """Return the Mie coefficients to compute the internal field of a sphere.

    Args:
        tau         integer: spherical polarization, 0 for spherical TE and 1 for spherical TM
        l           integer: l=1,... multipole degree (polar quantum number)
        k_medium    float or complex: wavenumber in surrounding medium (inverse length unit)
        k_particle  float or complex: wavenumber inside sphere (inverse length unit)
        radius      float: radius of sphere (length unit)

    Returns:
        Internal Mie coefficients as complex
    """
    jlkr_medium = sf.spherical_bessel(l, k_medium * radius)
    jlkr_particle = sf.spherical_bessel(l, k_particle * radius)
    dxxj_medium = sf.dx_xj(l, k_medium * radius)
    dxxj_particle = sf.dx_xj(l, k_particle * radius)

    hlkr_medium = sf.spherical_hankel(l, k_medium * radius)
    dxxh_medium = sf.dx_xh(l, k_medium * radius)

    if tau == 0:
        q = (jlkr_medium * dxxh_medium - hlkr_medium * dxxj_medium) / (jlkr_particle * dxxh_medium
                                                                       - hlkr_medium * dxxj_particle)
    elif tau == 1:
        q = ((k_medium * k_particle * jlkr_medium * dxxh_medium - k_medium * k_particle * hlkr_medium * dxxj_medium) /
             (k_particle ** 2 * jlkr_particle * dxxh_medium - k_medium ** 2 * hlkr_medium * dxxj_particle))
    else:
        raise ValueError('tau must be 0 (spherical TE) or 1 (spherical TM)')

    return q


def t_matrix_sphere(k_medium, k_particle, radius, l_max, m_max):
    """T-matrix of a spherical scattering object.

    Args:
        k_medium (float or complex):            Wavenumber in surrounding medium (inverse length unit)
        k_particle (float or complex):          Wavenumber inside sphere (inverse length unit)
        radius (float):                         Radius of sphere (length unit)
        l_max (int):                            Maximal multipole degree
        m_max (int):                            Maximal multipole order

    Returns:
         T-matrix as ndarray
    """
    t = np.zeros((flds.blocksize(l_max, m_max), flds.blocksize(l_max, m_max)), dtype=complex)
    for tau in range(2):
        for m in range(-m_max, m_max + 1):
            for l in range(max(1, abs(m)), l_max + 1):
                n = flds.multi_to_single_index(tau, l, m, l_max, m_max)
                t[n, n] = mie_coefficient(tau, l, k_medium, k_particle, radius)
    return t


def rotate_t_matrix(T, l_max, m_max, euler_angles, wdsympy=False):
    """T-matrix of a rotated particle.

    Args:
        T (numpy.array):        T-matrix
        l_max (int):            Maximal multipole degree
        m_max (int):            Maximal multipole order
        euler_angles (list):    Euler angles [alpha, beta, gamma] of rotated particle in (zy'z''-convention) in radian

    Returns:
        rotated T-matrix (numpy.array)
    """

    if euler_angles == [0, 0, 0]:
        return T
    else:
        # Doicu, Light Scattering by Systems of Particles, p. 70 (1.115)
        rot_mat_1 = trf.block_rotation_matrix_D_svwf(l_max, m_max, -euler_angles[2], -euler_angles[1],
                                                     -euler_angles[0], wdsympy)
        rot_mat_2 = trf.block_rotation_matrix_D_svwf(l_max, m_max, euler_angles[0], euler_angles[1], euler_angles[2],
                                                     wdsympy)
        T_mat_rot = (np.dot(np.dot(np.transpose(rot_mat_1), T), np.transpose(rot_mat_2)))

        # Mishchenko, Scattering, Absorption and Emission of Light by small Particles, p.120 (5.29)
        #       T_rot_matrix = np.dot(np.dot(trf.rotation_matrix_D(l_max, alpha, beta, gamma), T),
        #                             trf.rotation_matrix_D(l_max, -gamma, -beta, -alpha))

        return T_mat_rot




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  5 10:46:08 2022

This code is used to make the .c files and compile the Cython code from the
.pyx file. Note: The user will use the "compile_direct_coupling_block_from_c.c"
file, NOT this file. 


@author: parkerwray
"""


from setuptools import Extension, setup
from Cython.Build import cythonize
import numpy

ext_modules = [
    Extension(
        "cython_speedups",
        ["cython_speedups.pyx"],
        extra_compile_args=['-fopenmp'],
        extra_link_args=['-fopenmp'],
        include_dirs=[numpy.get_include()]
    )
]

setup(
    name='cython_speedups',
    ext_modules=cythonize(ext_modules),
)



"""
To compile cython_speedups.pyx, do the following:
    1) conda env list
        - This list paths to enviroments. Find one that has Cython installed
    2) conda activate [env path]
        - Run python scripts with this env.
    3) python compile_speedups_from_pyx.py build_ext --inplace
        - compile the c code using python to direct the compilation process.

NOTE: The function will first compile the .pyx code to a .c code. The function 
will then compile the .c code, using a c compiler, to a .so that can be run. 
The .so file is saved in the "build" folder. This folder houses the operating system
specific executables. If you want, you can copy the .so file to the main location 
to make importing the code easier.  

"""


"""This module contains functions to compute the layer mediated particle 
coupling coefficients."""

import numpy as np
import scipy.special
from numba import complex128, int64, jit
import smuthi.fields
import smuthi.fields.transformations as trf
import smuthi.layers as lay
import smuthi.utility.math as sf


@jit(complex128(complex128[:], complex128[:]),
     nopython=True, cache=True, nogil=True)
def numba_trapz(y, x):
    out = 0.0 + 0.0j
    #TODO implement some (optional) advanced summation?
    #e.g. https://github.com/nschloe/accupy/blob/master/accupy/sums.py
    #or better Sum2  from https://doi.org/10.1137/030601818 (Algorithm 4.4)
    #Note, that this may need to have exact summation for x and y, and exact product.
    for i in range( len(y) - 2 ):
        out += (x[i+1]-x[i]) * (y[i+1] + y[i])/2.0
    return out


@jit((complex128[:], complex128[:,:,:],
      complex128[:,:,:,:],complex128[:,:,:], int64),
     nopython=True,cache=True
     ,nogil=True
     # , parallel=True
    )
def eval_BeLBe(BeLBe, BeL, B1, ejkz, n2):
    for k in range(len(BeLBe)):
        for iplmn2 in range(2):
            for pol in range(2):
                BeLBe[k] += BeL[pol, iplmn2, k] * B1[pol, iplmn2, n2, k] * ejkz[1, 1 - iplmn2, k]
        

def g_function(vacuum_wavelength, receiving_particle, emitting_particle, layer_system, k_parallel):
    """
    This function returns the function g_n(k_\rho) as defined in equation (16) of the paper "A quick way to approximate
    a Sommerfeld-Weyl_type Sommerfeld integral" by Chew (1988) for the Sommerfeld integral for the layer-mediated
    particle coupling (compare Amos Egel's dissertation, equation (3.46)).
    The purpose of this function is to allow for stationary phase approximation in the case of  large lateral distances.

    The use of this function is constrained to the important special case of particles above a (possibly layered)
    substrate.

    Args:
        vacuum_wavelength (float):                          Vacuum wavelength :math:`\lambda` (length unit)
        receiving_particle (smuthi.particles.Particle):     Particle that receives the scattered field (must be in top
                                                            layer)
        emitting_particle (smuthi.particles.Particle):      Particle that emits the scattered field (must be in top
                                                            layer)
        layer_system (smuthi.layers.LayerSystem):           Stratified medium in which the coupling takes place
        k_parallel (float):                                 In-plane wavenumber

    Returns:
        a tuple with the following items:
        - g: a matrix of g_function as entries (dimension is blocksize x blocksize)
        - rho: radial distance between particles
        - delta_z: z-argument of e^(ik_z z) term
        - bessel_order: matrix of positive numbers to be used as the order of Hankel functions
    """

    # angular frequency
    omega = smuthi.fields.angular_frequency(vacuum_wavelength)

    # index specs
    blocksize1 = smuthi.fields.blocksize(receiving_particle.l_max, receiving_particle.m_max)
    blocksize2 = smuthi.fields.blocksize(emitting_particle.l_max, emitting_particle.m_max)

    # cylindrical coordinates of relative position vectors
    rs1 = np.array(receiving_particle.position)
    rs2 = np.array(emitting_particle.position)
    rs2s1 = rs1 - rs2
    rhos2s1 = np.linalg.norm(rs2s1[0:2])
    phis2s1 = np.arctan2(rs2s1[1], rs2s1[0])
    is1 = layer_system.layer_number(rs1[2])
    ziss1 = rs1[2] - layer_system.reference_z(is1)
    is2 = layer_system.layer_number(rs2[2])
    ziss2 = rs2[2] - layer_system.reference_z(is2)

    # assert that both particles are in top layer (this function applies only to this special case)
    assert (is1 == (layer_system.number_of_layers() - 1) and is2 == (layer_system.number_of_layers() - 1))

    # wave numbers
    k = omega * layer_system.refractive_indices[is1]
    kz = smuthi.fields.k_z(k_parallel=k_parallel, k=k)

    L = []
    for pol in range(2):
        # layer response
        # (we pick the [0, 1]-component corresponding to downgoing excitation and upgoing response
        # - should be the only non-zero)
        L.append(lay.layersystem_response_matrix(pol, layer_system.thicknesses,
                                                 layer_system.refractive_indices,
                                                 k_parallel, omega, is2, is1)[0, 1])

    # transformation coefficients, indices: pol, n
    Bdag = np.zeros((2, blocksize1), dtype=complex)
    B = np.zeros((2, blocksize2), dtype=complex)

    # multipole order
    m_vec = [np.zeros(blocksize1, dtype=int), np.zeros(blocksize2, dtype=int)]

    for tau in range(2):
        for m in range(-receiving_particle.m_max, receiving_particle.m_max + 1):
            for l in range(max(1, abs(m)), receiving_particle.l_max + 1):
                n = smuthi.fields.multi_to_single_index(tau, l, m, receiving_particle.l_max, receiving_particle.m_max)
                m_vec[0][n] = m
                for pol in range(2):
                    Bdag[pol, n] = trf.transformation_coefficients_vwf(tau, l, m, pol, kp=k_parallel, kz=kz, dagger=True)

    for tau in range(2):
        for m in range(-emitting_particle.m_max, emitting_particle.m_max + 1):
            for l in range(max(1, abs(m)), emitting_particle.l_max + 1):
                n = smuthi.fields.multi_to_single_index(tau, l, m, emitting_particle.l_max, emitting_particle.m_max)
                m_vec[1][n] = m
                for pol in range(2):
                    B[pol, n] = trf.transformation_coefficients_vwf(tau, l, m, pol, kp=k_parallel, kz=-kz, dagger=False)

    # factor
    m2_minus_m1 = m_vec[1] - m_vec[0][np.newaxis].T
    wr_const = 4 * (1j) ** abs(m2_minus_m1) * np.exp(1j * m2_minus_m1 * phis2s1) / (kz * k)

    # result
    g = np.zeros((blocksize1, blocksize2), dtype=complex)
    for pol in range(2):
        g += L[pol] * Bdag[pol, :][np.newaxis].T * B[pol, :]
    g *= wr_const

    # bessel function order
    bessel_order = abs(m2_minus_m1)
    delta_z = ziss1 + ziss2

    return g, rhos2s1, delta_z, bessel_order


def layer_mediated_coupling_block_stat_phase_approx(vacuum_wavelength, receiving_particle, emitting_particle,
                                                    layer_system):
    """
    Compute the layer mediated coupling coefficients by means of the stationary phase approximation, as presented in
    "A quick way to approximate a Sommerfeld-Weyl_type Sommerfeld integral" by W.C. Chew (1988).

    The stationary phase approximation is expected to yield good results for particles with a large lateral distance.

    ********************************************************************************************************************
    Note: This function assumes that both particles (emitter and receiver) are located in the top layer of the layered
    medium. For other cases, this function does not apply.
    ********************************************************************************************************************

    Args:
        vacuum_wavelength (float):                          Vacuum wavelength :math:`\lambda` (length unit)
        receiving_particle (smuthi.particles.Particle):     Particle that receives the scattered field
        emitting_particle (smuthi.particles.Particle):      Particle that emits the scattered field
        layer_system (smuthi.layers.LayerSystem):           Stratified medium in which the coupling takes place

    Returns:
        Stationary phase approximation for the layer mediated coupling matrix block as numpy array.

    """

    # index specs
    lmax1 = receiving_particle.l_max
    mmax1 = receiving_particle.m_max
    lmax2 = emitting_particle.l_max
    mmax2 = emitting_particle.m_max
    blocksize1 = smuthi.fields.blocksize(lmax1, mmax1)
    blocksize2 = smuthi.fields.blocksize(lmax2, mmax2)

    # cylindrical coordinates of relative position vectors
    rs1 = np.array(receiving_particle.position)
    rs2 = np.array(emitting_particle.position)
    rs2s1 = rs1 - rs2
    rhos2s1 = np.linalg.norm(rs2s1[0:2])
    phis2s1 = np.arctan2(rs2s1[1], rs2s1[0])
    is1 = layer_system.layer_number(rs1[2])
    ziss1 = rs1[2] - layer_system.reference_z(is1)
    is2 = layer_system.layer_number(rs2[2])
    ziss2 = rs2[2] - layer_system.reference_z(is2)

    # assert that both particles are in top layer (this function applies only to this special case)
    assert (is1 == (layer_system.number_of_layers() - 1) and is2 == (layer_system.number_of_layers() - 1))

    # wavenumber
    omega = smuthi.fields.angular_frequency(vacuum_wavelength)
    k = omega * layer_system.refractive_indices[is1]

    # stationary phase point
    r = np.sqrt(rhos2s1**2 + (ziss1 + ziss2)**2)
    sin_theta = rhos2s1 / r
    kpar0 = sin_theta * k
    kz0 = k * np.sqrt(1 - sin_theta**2)

    # g function (see [Chew1988])
    g, _, _, bess_ord = g_function(vacuum_wavelength=vacuum_wavelength,
                                   receiving_particle=receiving_particle,
                                   emitting_particle=emitting_particle,
                                   layer_system=layer_system,
                                   k_parallel=kpar0)

    # stationary phase approximation
    hankel = np.zeros((blocksize1, blocksize2), dtype=complex)

    for i1 in range(blocksize1):
        for i2 in range(blocksize2):
            hankel[i1, i2] = scipy.special.hankel1(bess_ord[i1, i2], kpar0 * rhos2s1)

    wr0 = -1j * g * hankel / scipy.special.hankel1(0, kpar0 * rhos2s1) * kz0 * np.exp(1j * k * r) / r

    return wr0


def layer_mediated_coupling_block(vacuum_wavelength, receiving_particle, emitting_particle, layer_system,
                                  k_parallel='default', show_integrand=False):
    """Layer-system mediated particle coupling matrix :math:`W^R` for two particles. This routine is explicit, but slow.

    Args:
        vacuum_wavelength (float):                          Vacuum wavelength :math:`\lambda` (length unit)
        receiving_particle (smuthi.particles.Particle):     Particle that receives the scattered field
        emitting_particle (smuthi.particles.Particle):      Particle that emits the scattered field
        layer_system (smuthi.layers.LayerSystem):           Stratified medium in which the coupling takes place
        k_parallel (numpy ndarray):                         In-plane wavenumbers for Sommerfeld integral
                                                            If 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        show_integrand (bool):                              If True, the norm of the integrand is plotted.

    Returns:
        Layer mediated coupling matrix block as numpy array.
    """
    if type(k_parallel) == str and k_parallel == 'default':
        k_parallel = smuthi.fields.default_Sommerfeld_k_parallel_array
       
    omega = smuthi.fields.angular_frequency(vacuum_wavelength)

    # index specs
    lmax1 = receiving_particle.l_max
    mmax1 = receiving_particle.m_max
    lmax2 = emitting_particle.l_max
    mmax2 = emitting_particle.m_max
    blocksize1 = smuthi.fields.blocksize(lmax1, mmax1)
    blocksize2 = smuthi.fields.blocksize(lmax2, mmax2)

    # cylindrical coordinates of relative position vectors
    rs1 = np.array(receiving_particle.position)
    rs2 = np.array(emitting_particle.position)
    rs2s1 = rs1 - rs2
    rhos2s1 = np.linalg.norm(rs2s1[0:2])
    phis2s1 = np.arctan2(rs2s1[1], rs2s1[0])
    is1 = layer_system.layer_number(rs1[2])
    ziss1 = rs1[2] - layer_system.reference_z(is1)
    is2 = layer_system.layer_number(rs2[2])
    ziss2 = rs2[2] - layer_system.reference_z(is2)

    # wave numbers
    kis1 = omega * layer_system.refractive_indices[is1]
    kis2 = omega * layer_system.refractive_indices[is2]
    kzis1 = smuthi.fields.k_z(k_parallel=k_parallel, k=kis1)
    kzis2 = smuthi.fields.k_z(k_parallel=k_parallel, k=kis2)

    # phase factors
    ejkz = np.zeros((2, 2, len(k_parallel)), dtype=complex)  # indices are: particle, plus/minus, kpar_idx
    ejkz[0, 0, :] = np.exp(1j * kzis1 * ziss1)
    ejkz[0, 1, :] = np.exp(- 1j * kzis1 * ziss1)
    ejkz[1, 0, :] = np.exp(1j * kzis2 * ziss2)
    ejkz[1, 1, :] = np.exp(- 1j * kzis2 * ziss2)

    # layer response
    L = np.zeros((2, 2, 2, len(k_parallel)), dtype=complex)  # polarization, pl/mn1, pl/mn2, kpar_idx
    for pol in range(2):
        L[pol, :, :, :] = lay.layersystem_response_matrix(pol, layer_system.thicknesses,
                                                          layer_system.refractive_indices, k_parallel, omega, is2, is1)

    # transformation coefficients
    B = [np.zeros((2, 2, blocksize1, len(k_parallel)), dtype=complex),
         np.zeros((2, 2, blocksize2, len(k_parallel)), dtype=complex)]
    # list index: particle, np indices: pol, plus/minus, n, kpar_idx

    m_vec = [np.zeros(blocksize1, dtype=int), np.zeros(blocksize2, dtype=int)]
   
    # precompute spherical functions
    ct = kzis1 / kis1
    st = k_parallel / kis1
    _, pilm_list_pl, taulm_list_pl = sf.legendre_normalized(ct, st, lmax1)
    _, pilm_list_mn, taulm_list_mn = sf.legendre_normalized(-ct, st, lmax1)
    pilm = (pilm_list_pl, pilm_list_mn)
    taulm = (taulm_list_pl, taulm_list_mn)
   
    for tau in range(2):
        for m in range(-mmax1, mmax1 + 1):
            for l in range(max(1, abs(m)), lmax1 + 1):
                n = smuthi.fields.multi_to_single_index(tau, l, m, lmax1, mmax1)
                m_vec[0][n] = m
                for iplmn in range(2):
                    for pol in range(2):
                        B[0][pol, iplmn, n, :] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm[iplmn],
                                                                                     taulm_list=taulm[iplmn], dagger=True)
   
    ct = kzis2 / kis2
    st = k_parallel / kis2
    _, pilm_list_pl, taulm_list_pl = sf.legendre_normalized(ct, st, lmax2)
    _, pilm_list_mn, taulm_list_mn = sf.legendre_normalized(-ct, st, lmax2)
    pilm = (pilm_list_pl, pilm_list_mn)
    taulm = (taulm_list_pl, taulm_list_mn)
                       
    for tau in range(2):
        for m in range(-mmax2, mmax2 + 1):
            for l in range(max(1, abs(m)), lmax2 + 1):
                n = smuthi.fields.multi_to_single_index(tau, l, m, lmax2, mmax2)
                m_vec[1][n] = m
                for iplmn in range(2):
                    for pol in range(2):
                        B[1][pol, iplmn, n, :] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm[iplmn],
                                                                                     taulm_list=taulm[iplmn], dagger=False)

    # bessel function and jacobi factor
    bessel_list = []
    for dm in range(lmax1 + lmax2 + 1):
        bessel_list.append(scipy.special.jv(dm, k_parallel * rhos2s1))
    jacobi_vector = k_parallel / (kzis2 * kis2)
    m2_minus_m1 = m_vec[1] - m_vec[0][np.newaxis].T
    wr_const = 4 * (1j) ** abs(m2_minus_m1) * np.exp(1j * m2_minus_m1 * phis2s1) 

    integral = np.zeros((blocksize1, blocksize2), dtype=complex) 
    for n1 in range(blocksize1):
        BeL = np.zeros((2, 2, len(k_parallel)), dtype=complex)  # indices are: pol, plmn2, n1, kpar_idx
        for iplmn1 in range(2):
            for pol in range(2):
                BeL[pol, :, :] += (L[pol, iplmn1, :, :]
                                        * B[0][pol, iplmn1, n1, :]
                                        * ejkz[0, iplmn1, :])
        for n2 in range(blocksize2):
            bessel_full = bessel_list[abs(m_vec[0][n1] - m_vec[1][n2])]
            BeLBe = np.zeros((len(k_parallel)), dtype=complex)
            eval_BeLBe(BeLBe, BeL, B[1], ejkz, n2)
            integrand = bessel_full * jacobi_vector * BeLBe
            integral[n1,n2] = numba_trapz(integrand, k_parallel)
    wr = wr_const * integral

    return wr


def layer_mediated_coupling_matrix(vacuum_wavelength, particle_list, layer_system, k_parallel='default'):
    """Layer system mediated particle coupling matrix W^R for a particle collection in a layered medium.

    Args:
        vacuum_wavelength (float):                                  Wavelength in length unit
        particle_list (list of smuthi.particles.Particle obejcts:   Scattering particles
        layer_system (smuthi.layers.LayerSystem):                   The stratified medium
        k_parallel (numpy.ndarray or str):                          In-plane wavenumber for Sommerfeld integrals.
                                                                    If 'default', smuthi.fields.default_Sommerfeld_k_parallel_array
   
    Returns:
        Ensemble coupling matrix as numpy array.
    """
   
    # indices
    blocksizes = [smuthi.fields.blocksize(particle.l_max, particle.m_max) for particle in particle_list]

    # initialize result
    wr = np.zeros((sum(blocksizes), sum(blocksizes)), dtype=complex)

    for s1, particle1 in enumerate(particle_list):
        idx1 = np.array(range(sum(blocksizes[:s1]), sum(blocksizes[:s1]) + blocksizes[s1]))
        for s2, particle2 in enumerate(particle_list):
            idx2 = range(sum(blocksizes[:s2]), sum(blocksizes[:s2]) + blocksizes[s2])
            wr[idx1[:, None], idx2] = layer_mediated_coupling_block(vacuum_wavelength, particle1, particle2,
                                                                    layer_system, k_parallel)

    return wr
